"""Manage steps to DGLAP solution, and operator creation."""
import os
from typing import Union

from ..io.runcards import OperatorCard, TheoryCard
from ..io.types import RawCard
from . import legacy


# TODO: drop this altogether, replacing just with managed.solve
# it is currently kept not to break the interface, but the runcards upgrade and
# path conversion should be done by the caller, here we just clearly declare
# which types we expect
def solve(
    theory_card: Union[RawCard, TheoryCard],
    operators_card: Union[RawCard, OperatorCard],
    path: os.PathLike,
):
    r"""Solve DGLAP equations in terms of evolution kernel operators (EKO).

    The EKO :math:`\mathbf E_{k,j}(a_s^1\leftarrow a_s^0)` is determined in order
    to fullfill the following evolution

    .. math::
        \mathbf f(x_k,a_s^1) = \mathbf E_{k,j}(a_s^1\leftarrow a_s^0) \mathbf f(x_j,a_s^0)

    The configuration is split between the theory settings, representing
    Standard Model parameters and other defining features of the theory
    calculation, and the operator settings, those that are more closely related
    to the solution of the |DGLAP| equation itself, and determine the resulting
    operator features.

    Note
    ----
    For further information about EKO inputs and output see :doc:`/code/IO`

    """
    # TODO: drop this
    legacy.Runner(theory_card, operators_card, path).compute()



"""Compute the leading-order Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko import constants

from ....harmonics import cache as c


@nb.njit(cache=True)
def gamma_ns(N, cache):
    r"""Compute the leading-order non-singlet anomalous dimension.

    Implements Eq. (3.4) of :cite:`Moch:2004pa`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache


    Returns
    -------
    gamma_ns : complex
        Leading-order non-singlet anomalous dimension :math:`\\gamma_{ns}^{(0)}(N)`
    """
    S1 = c.get(c.S1, cache, N)
    gamma = -(3.0 - 4.0 * S1 + 2.0 / N / (N + 1.0))
    result = constants.CF * gamma
    return result


@nb.njit(cache=True)
def gamma_qg(N, nf):
    r"""Compute the leading-order quark-gluon anomalous dimension.

    Implements Eq. (3.5) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors

    Returns
    -------
    gamma_qg : complex
        Leading-order quark-gluon anomalous dimension :math:`\\gamma_{qg}^{(0)}(N)`
    """
    gamma = -(N**2 + N + 2.0) / (N * (N + 1.0) * (N + 2.0))
    result = 2.0 * constants.TR * 2.0 * nf * gamma
    return result


@nb.njit(cache=True)
def gamma_gq(N):
    r"""Compute the leading-order gluon-quark anomalous dimension.

    Implements Eq. (3.5) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    gamma_gq : complex
        Leading-order gluon-quark anomalous dimension :math:`\\gamma_{gq}^{(0)}(N)`
    """
    gamma = -(N**2 + N + 2.0) / (N * (N + 1.0) * (N - 1.0))
    result = 2.0 * constants.CF * gamma
    return result


@nb.njit(cache=True)
def gamma_gg(N, cache, nf):
    r"""Compute the leading-order gluon-gluon anomalous dimension.

    Implements Eq. (3.5) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    nf : int
        Number of active flavors

    Returns
    -------
    gamma_gg : complex
        Leading-order gluon-gluon anomalous dimension :math:`\\gamma_{gg}^{(0)}(N)`
    """
    S1 = c.get(c.S1, cache, N)
    gamma = S1 - 1.0 / N / (N - 1.0) - 1.0 / (N + 1.0) / (N + 2.0)
    result = constants.CA * (4.0 * gamma - 11.0 / 3.0) + 4.0 / 3.0 * constants.TR * nf
    return result


@nb.njit(cache=True)
def gamma_singlet(N, cache, nf):
    r"""Compute the leading-order singlet anomalous dimension matrix.

    .. math::
        \\gamma_S^{(0)} = \\left(\begin{array}{cc}
        \\gamma_{qq}^{(0)} & \\gamma_{qg}^{(0)}\\
        \\gamma_{gq}^{(0)} & \\gamma_{gg}^{(0)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    nf : int
        Number of active flavors

    Returns
    -------
    gamma_S_0 : numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(0)}(N)`

    """
    gamma_qq = gamma_ns(N, cache)
    gamma_S_0 = np.array(
        [[gamma_qq, gamma_qg(N, nf)], [gamma_gq(N), gamma_gg(N, cache, nf)]],
        np.complex_,
    )
    return gamma_S_0


@nb.njit(cache=True)
def gamma_singlet_qed(N, cache, nf):
    r"""Compute the leading-order singlet anomalous dimension matrix for the unified evolution basis.

    .. math::
        \\gamma_S^{(1,0)} = \\left(\begin{array}{cccc}
        \\gamma_{gg}^{(1,0)} & 0 & \\gamma_{gq}^{(1,0)} & 0\\
        0 & 0 & 0 & 0 \\
        \\gamma_{qg}^{(1,0)} & 0 & \\gamma_{qq}^{(1,0)} & 0 \\
        0 & 0 & 0 & \\gamma_{qq}^{(1,0)} \\
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    nf : int
        Number of active flavors

    Returns
    -------
    gamma_S : numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(1,0)}(N)`

    """
    gamma_qq = gamma_ns(N, cache)
    gamma_S = np.array(
        [
            [gamma_gg(N, cache, nf), 0.0 + 0.0j, gamma_gq(N), 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [gamma_qg(N, nf), 0.0 + 0.0j, gamma_qq, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, gamma_qq],
        ],
        np.complex_,
    )
    return gamma_S


@nb.njit(cache=True)
def gamma_valence_qed(N, cache):
    r"""Compute the leading-order valence anomalous dimension matrix for the unified evolution basis.

    .. math::
        \\gamma_V^{(1,0)} = \\left(\begin{array}{cc}
        \\gamma_{ns}^{(1,0)} & 0\\
        0 & \\gamma_{ns}^{(1,0)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    gamma_V : numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{V}^{(1,0)}(N)`

    """
    gamma_V = np.array(
        [
            [1.0, 0.0],
            [0.0, 1.0],
        ],
        np.complex_,
    )
    return gamma_V * gamma_ns(N, cache)



"""The unpolarized, time-like |LO| Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko import constants

from ....harmonics import cache as c


@nb.njit(cache=True)
def gamma_qq(N, cache):
    r"""Compute the |LO| quark-quark anomalous dimension.

    Implements :eqref:`B.3` from :cite:`Mitov:2006wy`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        LO quark-quark anomalous dimension
        :math:`\gamma_{qq}^{(0)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    result = constants.CF * (-3.0 + (4.0 * S1) - 2.0 / (N * (N + 1.0)))
    return result


@nb.njit(cache=True)
def gamma_qg(N):
    r"""Compute the |LO| quark-gluon anomalous dimension.

    Implements :eqref:`B.4` from :cite:`Mitov:2006wy`
    and :eqref:`A1` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    complex
        LO quark-gluon anomalous dimension
        :math:`\gamma_{qg}^{(0)}(N)`

    """
    result = -(N**2 + N + 2.0) / (N * (N + 1.0) * (N + 2.0))
    return result


@nb.njit(cache=True)
def gamma_gq(N, nf):
    r"""Compute the |LO| gluon-quark anomalous dimension.

    Implements :eqref:`B.5` from :cite:`Mitov:2006wy`
    and :eqref:`A1` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors

    Returns
    -------
    complex
        LO quark-gluon anomalous dimension
        :math:`\gamma_{gq}^{(0)}(N)`

    """
    result = -4.0 * nf * constants.CF * (N**2 + N + 2.0) / (N * (N - 1.0) * (N + 1.0))
    return result


@nb.njit(cache=True)
def gamma_gg(N, nf, cache):
    r"""Compute the |LO| gluon-gluon anomalous dimension.

    Implements :eqref:`B.6` from :cite:`Mitov:2006wy`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        LO quark-quark anomalous dimension
        :math:`\gamma_{gg}^{(0)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    result = (2.0 * nf - 11.0 * constants.CA) / 3.0 + 4.0 * constants.CA * (
        S1 - 1.0 / (N * (N - 1.0)) - 1.0 / ((N + 1.0) * (N + 2.0))
    )
    return result


@nb.njit(cache=True)
def gamma_ns(N, cache):
    r"""Compute the |LO| non-singlet anomalous dimension.

    At LO, :math:`\gamma_{ns}^{(0)} = \gamma_{qq}^{(0)}`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        LO quark-quark anomalous dimension
        :math:`\gamma_{ns}^{(0)}(N)`

    """
    return gamma_qq(N, cache)


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache):
    r"""Compute the |LO| singlet anomalous dimension matrix.

    Implements :eqref:`2.13` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        LO singlet anomalous dimension matrix
        :math:`\gamma_{s}^{(0)}`
    """
    result = np.array(
        [
            [gamma_qq(N, cache), gamma_gq(N, nf)],
            [gamma_qg(N), gamma_gg(N, nf, cache)],
        ],
        np.complex_,
    )
    return result



"""Support legacy storage formats."""
import dataclasses
import io
import os
import pathlib
import tarfile
import tempfile
import warnings
from typing import Dict, List

import lz4.frame
import numpy as np
import yaml

from eko.interpolation import XGrid
from eko.io.runcards import flavored_mugrid
from eko.quantities.heavy_quarks import HeavyInfo, HeavyQuarkMasses, MatchingRatios

from . import raw
from .dictlike import DictLike
from .struct import EKO, Operator
from .types import EvolutionPoint as EPoint
from .types import RawCard


def load_tar(source: os.PathLike, dest: os.PathLike, errors: bool = False):
    """Load tar representation from file.

    Compliant with :meth:`dump_tar` output.

    Parameters
    ----------
    source :
        source tar name
    dest :
        dest tar name
    errors :
        whether to load also errors (default ``False``)

    """
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(tmpdir)

        with tarfile.open(source, "r") as tar:
            raw.safe_extractall(tar, tmpdir)

        # load metadata
        innerdir = list(tmpdir.glob("*"))[0]
        metapath = innerdir / "metadata.yaml"
        metaold = yaml.safe_load(metapath.read_text(encoding="utf-8"))

        theory = PseudoTheory.from_old(metaold)
        operator = PseudoOperator.from_old(metaold)

        # get actual grids
        arrays = load_arrays(innerdir)

    op5 = metaold.get("Q2grid")
    if op5 is None:
        op5 = metaold["mu2grid"]
    grid = op5to4(
        flavored_mugrid(op5, theory.heavy.masses, theory.heavy.matching_ratios), arrays
    )

    with EKO.create(dest) as builder:
        # here I'm plainly ignoring the static analyzer, the types are faking
        # the actual ones - not sure if I should fix builder interface to
        # accept also these
        eko = builder.load_cards(theory, operator).build()  # pylint: disable=E1101

        for ep, op in grid.items():
            eko[ep] = op

        eko.metadata.version = metaold.get("eko_version", "")
        eko.metadata.data_version = 0
        eko.metadata.update()


@dataclasses.dataclass
class PseudoTheory(DictLike):
    """Fake theory, mocking :class:`eko.io.runcards.TheoryCard`.

    Used to provide a theory for the :class:`~eko.io.struct.EKO` builder, even when the theory
    information is not available.

    """

    heavy: HeavyInfo

    @classmethod
    def from_old(cls, old: RawCard):
        """Load from old metadata."""
        heavy = HeavyInfo(
            num_flavs_init=4,
            num_flavs_max_pdf=None,
            intrinsic_flavors=None,
            masses=HeavyQuarkMasses([1.51, 4.92, 172.5]),
            masses_scheme=None,
            matching_ratios=MatchingRatios([1.0, 1.0, 1.0]),
        )
        return cls(heavy=heavy)


@dataclasses.dataclass
class PseudoOperator(DictLike):
    """Fake operator, mocking :class:`eko.io.runcards.OperatorCard`.

    Used to provide a theory for the :class:`~eko.io.struct.EKO` builder, even when the operator
    information is not fully available.

    """

    mu20: float
    evolgrid: List[EPoint]
    xgrid: XGrid
    configs: dict

    @classmethod
    def from_old(cls, old: RawCard):
        """Load from old metadata."""
        mu20 = float(old["q2_ref"])
        mu2list = old.get("Q2grid")
        if mu2list is None:
            mu2list = old["mu2grid"]
        mu2grid = np.array(mu2list)
        evolgrid = flavored_mugrid(
            np.sqrt(mu2grid).tolist(), [1.51, 4.92, 172.5], [1, 1, 1]
        )

        xgrid = XGrid(old["interpolation_xgrid"])

        configs = dict(
            interpolation_polynomial_degree=old.get("interpolation_polynomial_degree"),
            interpolation_is_log=old.get("interpolation_is_log"),
        )

        return cls(mu20=mu20, evolgrid=evolgrid, xgrid=xgrid, configs=configs)


ARRAY_SUFFIX = ".npy.lz4"
"""Suffix for array files inside the tarball."""


def load_arrays(dir: pathlib.Path) -> dict:
    """Load arrays from compressed dumps."""
    arrays = {}
    for fp in dir.glob(f"*{ARRAY_SUFFIX}"):
        with lz4.frame.open(fp, "rb") as fd:
            # static analyzer can not guarantee the content to be bytes
            content = fd.read()
            assert not isinstance(content, str), "Bytes expected"

            stream = io.BytesIO(content)
            stream.seek(0)
            arrays[pathlib.Path(fp.stem).stem] = np.load(stream)

    return arrays


OPERATOR = "operator"
"""File name stem for operators."""
ERROR = "operator_error"
"""File name stem for operator errrors."""


def op5to4(evolgrid: List[EPoint], arrays: dict) -> Dict[EPoint, Operator]:
    """Load dictionary of 4-dim operators, from a single 5-dim one."""

    def plural(name: str) -> list:
        return [name, f"{name}s"]

    op5 = None  # 5 dimensional operator
    err5 = None
    for name, ar in arrays.items():
        if name in plural(OPERATOR):
            op5 = ar
        elif name in plural(ERROR):
            err5 = ar
        else:
            warnings.warn(f"Unrecognized array loaded, '{name}'")

    if op5 is None:
        raise RuntimeError("Operator not found")
    if err5 is None:
        err5 = [None] * len(op5)

    grid = {}
    for ep, op4, err4 in zip(evolgrid, op5, err5):
        grid[ep] = Operator(operator=op4, error=err4)

    return grid



"""Base command definition."""
import click

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
def command():
    """Manage EKO assets and launch calculations."""



"""Weight 1 harmonic sums."""

import numba as nb
import numpy as np

from .polygamma import cern_polygamma


@nb.njit(cache=True)
def S1(N):
    r"""Compute the harmonic sum :math:`S_1(N)`.

    .. math::
      S_1(N) = \sum\limits_{j=1}^N \frac 1 j = \psi_0(N+1)+\gamma_E

    with :math:`\psi_0(N)` the digamma function and :math:`\gamma_E` the
    Euler-Mascheroni constant.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    S_1 : complex
        (simple) Harmonic sum :math:`S_1(N)`

    See Also
    --------
        ekore.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`

    """
    return cern_polygamma(N + 1.0, 0) + np.euler_gamma


@nb.njit(cache=True)
def Sm1(N, hS1, hS1mh, hS1h, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-1}(N)`.

    .. math::
      S_{-1}(N) = \sum\limits_{j=1}^N \frac {(-1)^j} j

    Parameters
    ----------
    N : complex
        Mellin moment
    hS1:  complex
        Harmonic sum :math:`S_{1}(N)`
    hS1mh: complex
        Harmonic sum :math:`S_{1}((N-1)/2)`
    hS1h: complex
        Harmonic sum :math:`S_{1}(N/2)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
        False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm1 : complex
        Harmonic sum :math:`S_{-1}(N)`

    See Also
    --------
        eko.anomalous_dimension.w1.S1 : :math:`S_1(N)`

    """
    if is_singlet is None:
        return (1 - (-1) ** N) / 2 * hS1mh + ((-1) ** N + 1) / 2 * hS1h - hS1
    if is_singlet:
        return hS1h - hS1
    return hS1mh - hS1



"""Main application class of eko."""
import logging
import os
from typing import Union

from ..evolution_operator.grid import OperatorGrid
from ..io import EKO, Operator, runcards
from ..io.types import RawCard
from . import commons

logger = logging.getLogger(__name__)


class Runner:
    """Represents a single input configuration.

    For details about the configuration, see :doc:`here </code/IO>`

    Attributes
    ----------
    setup : dict
        input configurations

    """

    banner = commons.BANNER

    def __init__(
        self,
        theory_card: Union[RawCard, runcards.TheoryCard],
        operators_card: Union[RawCard, runcards.OperatorCard],
        path: os.PathLike,
    ):
        """Initialize runner.

        Parameters
        ----------
        theory_card :
            theory parameters and options
        operators_card :
            operator specific options
        path :
            path where to store the computed operator

        """
        new_theory, new_operator = runcards.update(theory_card, operators_card)
        new_theory.heavy.intrinsic_flavors = [4, 5, 6]

        # Store inputs
        self.path = path
        self._theory = new_theory

        # setup basis grid
        bfd = commons.interpolator(new_operator)

        # call explicitly iter to explain the static analyzer that is an
        # iterable
        tc = commons.atlas(new_theory, new_operator)

        # strong coupling
        cs = commons.couplings(new_theory, new_operator)  # setup operator grid

        # compute masses if required
        masses = runcards.masses(new_theory, new_operator.configs.evolution_method)

        self.op_grid = OperatorGrid(
            mu2grid=new_operator.evolgrid,
            order=new_theory.order,
            masses=masses,
            mass_scheme=new_theory.heavy.masses_scheme.value,
            thresholds_ratios=new_theory.heavy.squared_ratios,
            intrinsic_flavors=new_theory.heavy.intrinsic_flavors,
            xif=new_theory.xif,
            configs=new_operator.configs,
            debug=new_operator.debug,
            atlas=tc,
            couplings=cs,
            interpol_dispatcher=bfd,
            n3lo_ad_variation=new_theory.n3lo_ad_variation,
        )

        with EKO.create(path) as builder:
            builder.load_cards(  # pylint: disable=E1101
                new_theory, new_operator
            ).build()

    def compute(self):
        """Run evolution and generate output operator.

        Two steps are applied sequentially:

        1. evolution is performed, computing the evolution operator in an
           internal flavor and x-space basis
        2. bases manipulations specified in the runcard are applied

        """
        with EKO.edit(self.path) as eko:
            # add all operators
            for ep, op in self.op_grid.compute().items():
                eko[ep] = Operator(**op)



r"""The unpolarized, space-like Altarelli-Parisi splitting kernels.

Normalization is given by

.. math::
    \mathbf{P}(x) = \sum\limits_{j=0} a_s^{j+1} \mathbf P^{(j)}(x)

with :math:`a_s = \frac{\alpha_S(\mu^2)}{4\pi}`.
The 3-loop references for the non-singlet :cite:`Moch:2004pa`
and singlet :cite:`Vogt:2004mw` case contain also the lower
order results. The results are also determined in Mellin space in
terms of the anomalous dimensions (note the additional sign!)

.. math::
    \gamma(N) = - \mathcal{M}[\mathbf{P}(x)](N)
"""

import numba as nb
import numpy as np

from eko import constants

from ....harmonics import cache as c
from . import aem1, aem2, as1, as1aem1, as2, as3, as4


@nb.njit(cache=True)
def gamma_ns(order, mode, n, nf):
    r"""Compute the tower of the non-singlet anomalous dimensions.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    mode : 10201 | 10101 | 10200
        sector identifier
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        non-singlet anomalous dimensions

    """
    cache = c.reset()
    # now combine
    gamma_ns = np.zeros(order[0], np.complex_)
    gamma_ns[0] = as1.gamma_ns(n, cache)
    # NLO and beyond
    if order[0] >= 2:
        if mode == 10101:
            gamma_ns_1 = as2.gamma_nsp(n, nf, cache)
        # To fill the full valence vector in NNLO we need to add gamma_ns^1 explicitly here
        elif mode in [10201, 10200]:
            gamma_ns_1 = as2.gamma_nsm(n, nf, cache)
        else:
            raise NotImplementedError("Non-singlet sector is not implemented")
        gamma_ns[1] = gamma_ns_1
    # NNLO and beyond
    if order[0] >= 3:
        if mode == 10101:
            gamma_ns_2 = as3.gamma_nsp(n, nf, cache)
        elif mode == 10201:
            gamma_ns_2 = as3.gamma_nsm(n, nf, cache)
        elif mode == 10200:
            gamma_ns_2 = as3.gamma_nsv(n, nf, cache)
        gamma_ns[2] = gamma_ns_2
    # N3LO
    if order[0] >= 4:
        if mode == 10101:
            gamma_ns_3 = as4.gamma_nsp(n, nf, cache)
        elif mode == 10201:
            gamma_ns_3 = as4.gamma_nsm(n, nf, cache)
        elif mode == 10200:
            gamma_ns_3 = as4.gamma_nsv(n, nf, cache)
        gamma_ns[3] = gamma_ns_3
    return gamma_ns


@nb.njit(cache=True)
def gamma_singlet(order, n, nf, n3lo_ad_variation):
    r"""Compute the tower of the singlet anomalous dimensions matrices.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    n : complex
        Mellin variable
    nf : int
        Number of active flavors
    n3lo_ad_variation : tuple
        |N3LO| anomalous dimension variation ``(gg_var, gq_var, qg_var, qq_var)``

    Returns
    -------
    numpy.ndarray
        singlet anomalous dimensions matrices

    """
    cache = c.reset()
    gamma_s = np.zeros((order[0], 2, 2), np.complex_)
    gamma_s[0] = as1.gamma_singlet(n, cache, nf)
    if order[0] >= 2:
        gamma_s[1] = as2.gamma_singlet(n, nf, cache)
    if order[0] >= 3:
        gamma_s[2] = as3.gamma_singlet(n, nf, cache)
    if order[0] >= 4:
        gamma_s[3] = as4.gamma_singlet(n, nf, cache, n3lo_ad_variation)
    return gamma_s


@nb.njit(cache=True)
def gamma_ns_qed(order, mode, n, nf):
    r"""
    Compute the grid of the QED non-singlet anomalous dimensions.

    Parameters
    ----------
        order : tuple(int,int)
            perturbative orders
        mode : 10102 | 10103 | 10202 | 10203
            sector identifier
        n : complex
            Mellin variable
        nf : int
            Number of active flavors

    Returns
    -------
        gamma_ns : numpy.ndarray
            non-singlet QED anomalous dimensions
    """
    cache = c.reset()
    # now combine
    gamma_ns = np.zeros((order[0] + 1, order[1] + 1), np.complex_)
    gamma_ns[1, 0] = as1.gamma_ns(n, cache)
    gamma_ns[0, 1] = choose_ns_ad_aem1(mode, n, cache)
    gamma_ns[1, 1] = choose_ns_ad_as1aem1(mode, n, cache)
    # NLO and beyond
    if order[0] >= 2:
        if mode in [10102, 10103]:
            gamma_ns[2, 0] = as2.gamma_nsp(n, nf, cache)
        # To fill the full valence vector in NNLO we need to add gamma_ns^1 explicitly here
        else:
            gamma_ns[2, 0] = as2.gamma_nsm(n, nf, cache)
    if order[1] >= 2:
        gamma_ns[0, 2] = choose_ns_ad_aem2(mode, n, nf, cache)
    # NNLO and beyond
    if order[0] >= 3:
        if mode in [10102, 10103]:
            gamma_ns[3, 0] = as3.gamma_nsp(n, nf, cache)
        elif mode in [10202, 10203]:
            gamma_ns[3, 0] = as3.gamma_nsm(n, nf, cache)
    return gamma_ns


@nb.njit(cache=True)
def choose_ns_ad_aem1(mode, n, cache):
    r"""
    Select the non-singlet anomalous dimension at O(aem1) with the correct charge factor.

    Parameters
    ----------
    mode : 10102 | 10202 | 10103 | 10203
        sector identifier
    n : complex
        Mellin variable
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        non-singlet anomalous dimensions

    """
    if mode in [10102, 10202]:
        return constants.eu2 * aem1.gamma_ns(n, cache)
    elif mode in [10103, 10203]:
        return constants.ed2 * aem1.gamma_ns(n, cache)
    raise NotImplementedError("Non-singlet sector is not implemented")


@nb.njit(cache=True)
def choose_ns_ad_as1aem1(mode, n, cache):
    r"""
    Select the non-singlet anomalous dimension at O(as1aem1) with the correct charge factor.

    Parameters
    ----------
    mode : 10102 | 10202 | 10103 | 10203
        sector identifier
    n : complex
        Mellin variable
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        non-singlet anomalous dimensions

    """
    if mode == 10102:
        return constants.eu2 * as1aem1.gamma_nsp(n, cache)
    elif mode == 10103:
        return constants.ed2 * as1aem1.gamma_nsp(n, cache)
    elif mode == 10202:
        return constants.eu2 * as1aem1.gamma_nsm(n, cache)
    elif mode == 10203:
        return constants.ed2 * as1aem1.gamma_nsm(n, cache)
    raise NotImplementedError("Non-singlet sector is not implemented")


@nb.njit(cache=True)
def choose_ns_ad_aem2(mode, n, nf, cache):
    r"""
    Select the non-singlet anomalous dimension at O(aem2) with the correct charge factor.

    Parameters
    ----------
    mode : 10102 | 10202 | 10103 | 10203
        sector identifier
    n : complex
        Mellin variable
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        non-singlet anomalous dimensions

    """
    if mode == 10102:
        return constants.eu2 * aem2.gamma_nspu(n, nf, cache)
    elif mode == 10103:
        return constants.ed2 * aem2.gamma_nspd(n, nf, cache)
    elif mode == 10202:
        return constants.eu2 * aem2.gamma_nsmu(n, nf, cache)
    elif mode == 10203:
        return constants.ed2 * aem2.gamma_nsmd(n, nf, cache)
    raise NotImplementedError("Non-singlet sector is not implemented")


@nb.njit(cache=True)
def gamma_singlet_qed(order, n, nf):
    r"""
    Compute the grid of the QED singlet anomalous dimensions matrices.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        singlet anomalous dimensions matrices

    """
    cache = c.reset()
    gamma_s = np.zeros((order[0] + 1, order[1] + 1, 4, 4), np.complex_)
    gamma_s[1, 0] = as1.gamma_singlet_qed(n, cache, nf)
    gamma_s[0, 1] = aem1.gamma_singlet(n, nf, cache)
    gamma_s[1, 1] = as1aem1.gamma_singlet(n, nf, cache)
    if order[0] >= 2:
        gamma_s[2, 0] = as2.gamma_singlet_qed(n, nf, cache)
    if order[1] >= 2:
        gamma_s[0, 2] = aem2.gamma_singlet(n, nf, cache)
    if order[0] >= 3:
        gamma_s[3, 0] = as3.gamma_singlet_qed(n, nf, cache)
    return gamma_s


@nb.njit(cache=True)
def gamma_valence_qed(order, n, nf):
    r"""
    Compute the grid of the QED valence anomalous dimensions matrices.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        valence anomalous dimensions matrices

    """
    cache = c.reset()
    gamma_v = np.zeros((order[0] + 1, order[1] + 1, 2, 2), np.complex_)
    gamma_v[1, 0] = as1.gamma_valence_qed(n, cache)
    gamma_v[0, 1] = aem1.gamma_valence(n, nf, cache)
    gamma_v[1, 1] = as1aem1.gamma_valence(n, nf, cache)
    if order[0] >= 2:
        gamma_v[2, 0] = as2.gamma_valence_qed(n, nf, cache)
    if order[1] >= 2:
        gamma_v[0, 2] = aem2.gamma_valence(n, nf, cache)
    if order[0] >= 3:
        gamma_v[3, 0] = as3.gamma_valence_qed(n, nf, cache)
    return gamma_v



"""Runners common utilities."""
import numpy as np

from ..couplings import Couplings, couplings_mod_ev
from ..interpolation import InterpolatorDispatcher
from ..io import runcards
from ..io.runcards import OperatorCard, TheoryCard
from ..io.types import ScaleVariationsMethod
from ..matchings import Atlas

BANNER = r"""
oooooooooooo oooo    oooo  \\ .oooooo.
`888'     `8 `888   .8P'  //////    `Y8b
 888          888  d8'   \\o\/////    888
 888oooo8     88888     \\\\/8/////   888
 888    "     888`88b.      888 ///   888
 888       o  888  `88b.    `88b //  d88'
o888ooooood8 o888o  o888o     `Y8bood8P'
"""


def interpolator(operator: OperatorCard) -> InterpolatorDispatcher:
    """Create interpolator from runcards."""
    return InterpolatorDispatcher(
        xgrid=operator.xgrid,
        polynomial_degree=operator.configs.interpolation_polynomial_degree,
    )


def atlas(theory: TheoryCard, operator: OperatorCard) -> Atlas:
    """Create thresholds atlas from runcards."""
    # TODO: cache result
    masses = runcards.masses(theory, operator.configs.evolution_method)
    matching_scales = np.power(theory.heavy.matching_ratios, 2.0) * np.array(masses)
    return Atlas(matching_scales.tolist(), (operator.mu20, theory.heavy.num_flavs_init))


def couplings(theory: TheoryCard, operator: OperatorCard) -> Couplings:
    """Create couplings from runcards."""
    thresholds_ratios = np.power(theory.heavy.matching_ratios, 2.0)
    masses = runcards.masses(theory, operator.configs.evolution_method)
    return Couplings(
        couplings=theory.couplings,
        order=theory.order,
        method=couplings_mod_ev(operator.configs.evolution_method),
        masses=masses,
        hqm_scheme=theory.heavy.masses_scheme,
        thresholds_ratios=thresholds_ratios
        * (
            theory.xif**2
            if operator.configs.scvar_method == ScaleVariationsMethod.EXPONENTIATED
            else 1.0
        ),
    )



"""Contains the definitions of the :doc:`Flavor Basis and Evolution Basis </theory/FlavorSpace>`."""

import numpy as np

flavor_basis_pids = tuple([22] + list(range(-6, -1 + 1)) + [21] + list(range(1, 6 + 1)))
r"""
Sorted elements in Flavor Basis as |pid|.

Definition: `here <https://pdg.lbl.gov/2019/reviews/rpp2019-rev-monte-carlo-numbering.pdf>`_

corresponding |PDF| : :math:`\gamma, \bar t, \bar b, \bar c, \bar s, \bar u, \bar d, g,
d, u, s, c, b, t`
"""

flavor_basis_names = (
    "ph",
    "tbar",
    "bbar",
    "cbar",
    "sbar",
    "ubar",
    "dbar",
    "g",
    "d",
    "u",
    "s",
    "c",
    "b",
    "t",
)
"""String representation of :data:`flavor_basis_pids`."""


quark_names = "".join(flavor_basis_names[-6:])

evol_basis = (
    "ph",
    "S",
    "g",
    "V",
    "V3",
    "V8",
    "V15",
    "V24",
    "V35",
    "T3",
    "T8",
    "T15",
    "T24",
    "T35",
)
r"""
Sorted elements in Evolution Basis as :obj:`str`.

Definition: :ref:`here <theory/FlavorSpace:flavor basis>`.

corresponding |PDF| : :math:`\gamma, \Sigma, g, V, V_{3}, V_{8}, V_{15}, V_{24},
V_{35}, T_{3}, T_{8}, T_{15}, T_{24}, T_{35}`
"""
unified_evol_basis = (
    "g",
    "ph",
    "S",
    "Sdelta",
    "V",
    "Vdelta",
    "Td3",
    "Vd3",
    "Tu3",
    "Vu3",
    "Td8",
    "Vd8",
    "Tu8",
    "Vu8",
)
r"""
Sorted elements in Unified Evolution Basis as :obj:`str`.

Definition: :ref:`here <theory/FlavorSpace:flavor basis>`.

corresponding |PDF| : :math:`g, \gamma, \Sigma, \Sigma_\Delta, V, V_\Delta, T^d_3, V^d_3,
T^u_3, V^u_3, T^d_8, V^d_8, T^u_8, V^u_8`
"""

evol_basis_pids = tuple(
    [22, 100, 21, 200]
    + [200 + n**2 - 1 for n in range(2, 6 + 1)]
    + [100 + n**2 - 1 for n in range(2, 6 + 1)]
)
"""|pid| representation of :data:`evol_basis`."""

unified_evol_basis_pids = tuple(
    [21, 22, 100, 101, 200, 201]
    + [103 + 2, 203 + 2]  # d = +2
    + [103 + 1, 203 + 1]  # u = +1
    + [108 + 2, 208 + 2]
    + [108 + 1, 208 + 1]
)
r"""
|pid| representation of :data:`unified_evol_basis`.

The notation used for the non singlet compunents is the following:
pid_ns(u) = pid_ns + 1, pid_ns(d) = pid_ns + 2.
"""

non_singlet_pids_map = {
    "ns-": 10201,
    "ns+": 10101,
    "nsV": 10200,
    "ns-u": 10202,
    "ns-d": 10203,
    "ns+u": 10102,
    "ns+d": 10103,
}

singlet_labels = ((100, 100), (100, 21), (21, 100), (21, 21))
non_singlet_labels = (
    (non_singlet_pids_map["ns-"], 0),
    (non_singlet_pids_map["ns+"], 0),
    (non_singlet_pids_map["nsV"], 0),
)
singlet_unified_labels = (
    (21, 21),
    (21, 22),
    (21, 100),
    (21, 101),  # Sdelta = 101
    (22, 21),
    (22, 22),
    (22, 100),
    (22, 101),
    (100, 21),
    (100, 22),
    (100, 100),
    (100, 101),
    (101, 21),
    (101, 22),
    (101, 100),
    (101, 101),
)
valence_unified_labels = (
    (10200, 10200),
    (10200, 10204),  # Vdelta = 10204
    (10204, 10200),
    (10204, 10204),
)
non_singlet_unified_labels = (
    (non_singlet_pids_map["ns+d"], 0),
    (non_singlet_pids_map["ns-d"], 0),
    (non_singlet_pids_map["ns+u"], 0),
    (non_singlet_pids_map["ns-u"], 0),
)
full_labels = (*singlet_labels, *non_singlet_labels)
full_unified_labels = (
    *singlet_unified_labels,
    *valence_unified_labels,
    *non_singlet_unified_labels,
)

anomalous_dimensions_basis = full_labels
r"""
Sorted elements in Anomalous Dimensions Basis as :obj:`str`.
"""
matching_hplus_pid = 90
matching_hminus_pid = 91

# Tranformation from physical basis to QCD evolution basis
rotate_flavor_to_evolution = np.array(
    [
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -1, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, -0, -1, 1, 0, -1, 1, 0, -0, 0, 0],
        [0, 0, 0, 0, 2, -1, -1, 0, 1, 1, -2, -0, 0, 0],
        [0, 0, 0, 3, -1, -1, -1, 0, 1, 1, 1, -3, 0, 0],
        [0, 0, 4, -1, -1, -1, -1, 0, 1, 1, 1, 1, -4, 0],
        [0, 5, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, -5],
        [0, 0, 0, 0, 0, 1, -1, 0, -1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, -2, 1, 1, 0, 1, 1, -2, 0, 0, 0],
        [0, 0, 0, -3, 1, 1, 1, 0, 1, 1, 1, -3, 0, 0],
        [0, 0, -4, 1, 1, 1, 1, 0, 1, 1, 1, 1, -4, 0],
        [0, -5, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, -5],
    ]
)
"""
Basis rotation matrix between :doc:`Flavor Basis and Evolution Basis </theory/FlavorSpace>`.
"""

# Tranformation from physical basis to QCDxQED evolution basis
rotate_flavor_to_unified_evolution = np.array(
    [
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
        [0, 1, -1, 1, -1, 1, -1, 0, -1, 1, -1, 1, -1, 1],
        [0, -1, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, 1],
        [0, -1, 1, -1, 1, -1, 1, 0, -1, 1, -1, 1, -1, 1],
        [0, 0, 0, 0, -1, 0, 1, 0, 1, 0, -1, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, -1, 0, 1, 0, -1, 0, 0, 0],
        [0, 0, 0, -1, 0, 1, 0, 0, 0, 1, 0, -1, 0, 0],
        [0, 0, 0, 1, 0, -1, 0, 0, 0, 1, 0, -1, 0, 0],
        [0, 0, -2, 0, 1, 0, 1, 0, 1, 0, 1, 0, -2, 0],
        [0, 0, 2, 0, -1, 0, -1, 0, 1, 0, 1, 0, -2, 0],
        [0, -2, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, -2],
        [0, 2, 0, -1, 0, -1, 0, 0, 0, 1, 0, 1, 0, -2],
    ]
)
"""
Basis rotation matrix between :doc:`Flavor Basis and Unified Evolution Basis </theory/FlavorSpace>`.
"""


map_ad_to_evolution = {
    (100, 100): ["S.S"],
    (100, 21): ["S.g"],
    (21, 100): ["g.S"],
    (21, 21): ["g.g"],
    (non_singlet_pids_map["nsV"], 0): ["V.V"],
    (non_singlet_pids_map["ns+"], 0): [
        "T3.T3",
        "T8.T8",
        "T15.T15",
        "T24.T24",
        "T35.T35",
    ],
    (non_singlet_pids_map["ns-"], 0): [
        "V3.V3",
        "V8.V8",
        "V15.V15",
        "V24.V24",
        "V35.V35",
    ],
}
"""
Map anomalous dimension sectors' names to their members
"""

map_ad_to_unified_evolution = {
    (21, 21): ["g.g"],
    (21, 22): ["g.ph"],
    (21, 100): ["g.S"],
    (21, 101): ["g.Sdelta"],
    (22, 21): ["ph.g"],
    (22, 22): ["ph.ph"],
    (22, 100): ["ph.S"],
    (22, 101): ["ph.Sdelta"],
    (100, 21): ["S.g"],
    (100, 22): ["S.ph"],
    (100, 100): ["S.S"],
    (100, 101): ["S.Sdelta"],
    (101, 21): ["Sdelta.g"],
    (101, 22): ["Sdelta.ph"],
    (101, 100): ["Sdelta.S"],
    (101, 101): ["Sdelta.Sdelta"],
    (10200, 10200): ["V.V"],
    (10200, 10204): ["V.Vdelta"],
    (10204, 10200): ["Vdelta.V"],
    (10204, 10204): ["Vdelta.Vdelta"],
    (non_singlet_pids_map["ns+u"], 0): [
        "Tu3.Tu3",
        "Tu8.Tu8",
    ],
    (non_singlet_pids_map["ns+d"], 0): [
        "Td3.Td3",
        "Td8.Td8",
    ],
    (non_singlet_pids_map["ns-u"], 0): [
        "Vu3.Vu3",
        "Vu8.Vu8",
    ],
    (non_singlet_pids_map["ns-d"], 0): [
        "Vd3.Vd3",
        "Vd8.Vd8",
    ],
}


def ad_projector(ad_lab, nf, qed=False):
    """
    Build a projector (as a numpy array) for the given anomalous dimension sector.

    Parameters
    ----------
    ad_lab : str
        name of anomalous dimension sector
    nf : int
        number of light flavors
    qed : bool
        activate qed

    Returns
    -------
    proj : np.ndarray
        projector over the specified sector
    """
    if not qed:
        proj = np.zeros_like(rotate_flavor_to_evolution, dtype=float)
        # restrict the evolution basis to light flavors
        # NOTE: the cut is only needed for "NS_p" and "NS_m", but the other lists
        # are 1-long so they are unaffected
        sector = map_ad_to_evolution[ad_lab][: (nf - 1)]
        basis = evol_basis
        rotate = rotate_flavor_to_evolution.copy()
    else:
        proj = np.zeros_like(rotate_flavor_to_unified_evolution, dtype=float)
        # restrict the evolution basis to light flavors
        sector = select_light_flavors_uni_ev(ad_lab, nf)
        basis = unified_evol_basis
        rotate = rotate_flavor_to_unified_evolution.copy()

    for el in sector:
        out_name, in_name = el.split(".")
        out_idx = basis.index(out_name)
        in_idx = basis.index(in_name)
        out = rotate[out_idx].copy()
        in_ = rotate[in_idx].copy()
        out[: 1 + (6 - nf)] = out[len(out) - (6 - nf) :] = 0
        in_[: 1 + (6 - nf)] = in_[len(in_) - (6 - nf) :] = 0
        proj += (out[:, np.newaxis] * in_) / (out @ out)

    return proj


def select_light_flavors_uni_ev(ad_lab, nf):
    """
    Select light flavors for a given ad_lab.

    Parameters
    ----------
    ad_lab : str
        name of anomalous dimension sector
    nf : int
        number of light flavors

    Returns
    -------
    sector : list
        list of sector for a given ad_lab
    """
    if ad_lab[0] in [non_singlet_pids_map["ns+d"], non_singlet_pids_map["ns-d"]]:
        if nf < 5:
            # return only Td3 or Vd3
            return [map_ad_to_unified_evolution[ad_lab][0]]
        else:
            return map_ad_to_unified_evolution[ad_lab]
    elif ad_lab[0] in [non_singlet_pids_map["ns+u"], non_singlet_pids_map["ns-u"]]:
        if nf < 4:
            return []
        elif nf < 6 and nf >= 4:
            # return only Td3 or Vd3
            return [map_ad_to_unified_evolution[ad_lab][0]]
        else:
            return map_ad_to_unified_evolution[ad_lab]
    else:
        return map_ad_to_evolution[ad_lab]


def ad_projectors(nf, qed=False):
    """
    Build projectors tensor (as a numpy array), collecting all the individual sector projectors.

    Parameters
    ----------
    nf : int
        number of light flavors
    qed : bool
        activate qed

    Returns
    -------
    projs : np.ndarray
        projectors tensor
    """
    projs = []
    for ad in anomalous_dimensions_basis:
        projs.append(ad_projector(ad, nf, qed))

    return np.array(projs)


def intrinsic_unified_evol_labels(nf):
    """
    Collect all labels in the intrinsic unified evolution basis.

    Parameters
    ----------
    nf : int
        number of light flavors

    Returns
    -------
    labels : list(str)
        active distributions
    """
    labels = [
        "ph",
        "g",
        "S",
        "V",
        "Sdelta",
        "Vdelta",
        "Td3",
        "Vd3",
    ]
    news = ["u3", "d8", "u8"]
    for j in range(nf, 6):
        q = quark_names[j]
        labels.extend([f"{q}+", f"{q}-"])
    for k in range(3, nf):
        new = news[k - 3]
        labels.extend([f"T{new}", f"V{new}"])

    return labels



"""Manage file system resources access."""

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from . import exceptions


class ReadOnlyOperator(RuntimeError, exceptions.OutputError):
    """It is not possible to write on a read-only operator.

    In particular, the behavior would be deceitful, since writing is possible
    in-memory and even on the temporary folder.
    But eventually, no writing will happen on a persistent archive, so any
    modification is lost after exiting the program.

    """


class ClosedOperator(RuntimeError, exceptions.OutputError):
    """It is not possible to write on nor to read from a closed operator.

    This is milder issue than :class:`ReadOnlyOperator`, since in this case not
    even the writing on the temporary folder would be possible.

    Instead, it will look like you can access some properties, but the operator
    is actually closed, so it should not be used any longer in general.
    However, for extremely simple properties, like those available in memory
    from :class:`eko.io.struct.Metadata` or :class:`eko.io.struct.AccessConfigs`, there
    is no need to raise on read, since those properties are actually available,
    but they should always raise on writing, since there is no persistence for
    the content written, and it can be deceitful.

    Still, the level of protection will be mild, since a thoruough protection
    would clutter a lot the code, requiring a lot of maintenance.
    "We are adult here".

    """


@dataclass
class AccessConfigs:
    """Configurations specified during opening of an EKO."""

    path: Path
    """The path to the permanent object."""
    readonly: bool
    "Read-only flag"
    open: bool
    "EKO status"

    @property
    def read(self):
        """Check reading permission.

        Reading access is always granted on open operator.

        """
        return self.open

    @property
    def write(self):
        """Check writing permission."""
        return self.open and not self.readonly

    def assert_open(self):
        """Assert operator is open.

        Raises
        ------
        exceptions.ClosedOperator
            if operator is closed

        """
        if not self.open:
            raise ClosedOperator

    def assert_writeable(self, msg: Optional[str] = None):
        """Assert operator is writeable.

        Raises
        ------
        exceptions.ClosedOperator
            see :meth:`assert_open`
        exceptions.ReadOnlyOperator
            if operators has been declared read-only

        """
        if msg is None:
            msg = ""

        self.assert_open()
        if self.readonly:
            raise ReadOnlyOperator(msg)



r"""Contains the central operator classes.

See :doc:`Operator overview </code/Operators>`.
"""

import functools
import logging
import os
import time
from multiprocessing import Pool

import numba as nb
import numpy as np
from scipy import integrate

import ekore.anomalous_dimensions.polarized.space_like as ad_ps
import ekore.anomalous_dimensions.unpolarized.space_like as ad_us
import ekore.anomalous_dimensions.unpolarized.time_like as ad_ut

from .. import basis_rotation as br
from .. import interpolation, mellin
from .. import scale_variations as sv
from ..kernels import non_singlet as ns
from ..kernels import non_singlet_qed as qed_ns
from ..kernels import singlet as s
from ..kernels import singlet_qed as qed_s
from ..kernels import utils
from ..kernels import valence_qed as qed_v
from ..matchings import Segment
from ..member import OpMember

logger = logging.getLogger(__name__)


@nb.njit(cache=True)
def select_singlet_element(ker, mode0, mode1):
    """Select element of the singlet matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        singlet integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element

    Returns
    -------
    complex
        singlet integration kernel element
    """
    k = 0 if mode0 == 100 else 1
    l = 0 if mode1 == 100 else 1
    return ker[k, l]


@nb.njit(cache=True)
def select_QEDsinglet_element(ker, mode0, mode1):
    """Select element of the QEDsinglet matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        QEDsinglet integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element
    Returns
    -------
    ker : complex
        QEDsinglet integration kernel element
    """
    if mode0 == 21:
        index1 = 0
    elif mode0 == 22:
        index1 = 1
    elif mode0 == 100:
        index1 = 2
    else:
        index1 = 3
    if mode1 == 21:
        index2 = 0
    elif mode1 == 22:
        index2 = 1
    elif mode1 == 100:
        index2 = 2
    else:
        index2 = 3
    return ker[index1, index2]


@nb.njit(cache=True)
def select_QEDvalence_element(ker, mode0, mode1):
    """
    Select element of the QEDvalence matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        QEDvalence integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element
    Returns
    -------
    ker : complex
        QEDvalence integration kernel element
    """
    index1 = 0 if mode0 == 10200 else 1
    index2 = 0 if mode1 == 10200 else 1
    return ker[index1, index2]


spec = [
    ("is_singlet", nb.boolean),
    ("is_QEDsinglet", nb.boolean),
    ("is_QEDvalence", nb.boolean),
    ("is_log", nb.boolean),
    ("logx", nb.float64),
    ("u", nb.float64),
]


@nb.experimental.jitclass(spec)
class QuadKerBase:
    """Manage the common part of Mellin inversion integral.

    Parameters
    ----------
    u : float
        quad argument
    is_log : boolean
        is a logarithmic interpolation
    logx : float
        Mellin inversion point
    mode0 : str
        first sector element
    """

    def __init__(self, u, is_log, logx, mode0):
        self.is_singlet = mode0 in [100, 21, 90]
        self.is_QEDsinglet = mode0 in [21, 22, 100, 101, 90]
        self.is_QEDvalence = mode0 in [10200, 10204]
        self.is_log = is_log
        self.u = u
        self.logx = logx

    @property
    def path(self):
        """Return the associated instance of :class:`eko.mellin.Path`."""
        if self.is_singlet or self.is_QEDsinglet:
            return mellin.Path(self.u, self.logx, True)
        else:
            return mellin.Path(self.u, self.logx, False)

    @property
    def n(self):
        """Returs the Mellin moment :math:`N`."""
        return self.path.n

    def integrand(
        self,
        areas,
    ):
        """Get transformation to Mellin space integral.

        Parameters
        ----------
        areas : tuple
            basis function configuration

        Returns
        -------
        complex
            common mellin inversion integrand
        """
        if self.logx == 0.0:
            return 0.0
        pj = interpolation.evaluate_grid(self.path.n, self.is_log, self.logx, areas)
        if pj == 0.0:
            return 0.0
        return self.path.prefactor * pj * self.path.jac


@nb.njit(cache=True)
def quad_ker(
    u,
    order,
    mode0,
    mode1,
    method,
    is_log,
    logx,
    areas,
    as_list,
    mu2_from,
    mu2_to,
    a_half,
    alphaem_running,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
    n3lo_ad_variation,
    is_polarized,
    is_time_like,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    u : float
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    is_log : boolean
        is a logarithmic interpolation
    logx : float
        Mellin inversion point
    areas : tuple
        basis function configuration
    as1 : float
        target coupling value
    as0 : float
        initial coupling value
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2
    aem_list : list
        list of electromagnetic coupling values
    alphaem_running : bool
        whether alphaem is running or not
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an intermediate threshold operator?
    n3lo_ad_variation : tuple
        |N3LO| anomalous dimension variation ``(gg_var, gq_var, qg_var, qq_var)``
    is_polarized : boolean
        is polarized evolution ?
    is_time_like : boolean
        is time-like evolution ?

    Returns
    -------
    float
        evaluated integration kernel
    """
    ker_base = QuadKerBase(u, is_log, logx, mode0)
    integrand = ker_base.integrand(areas)
    if integrand == 0.0:
        return 0.0
    if order[1] == 0:
        ker = quad_ker_qcd(
            ker_base,
            order,
            mode0,
            mode1,
            method,
            as_list[-1],
            as_list[0],
            nf,
            L,
            ev_op_iterations,
            ev_op_max_order,
            sv_mode,
            is_threshold,
            is_polarized,
            is_time_like,
            n3lo_ad_variation,
        )
    else:
        ker = quad_ker_qed(
            ker_base,
            order,
            mode0,
            mode1,
            method,
            as_list,
            mu2_from,
            mu2_to,
            a_half,
            alphaem_running,
            nf,
            L,
            ev_op_iterations,
            ev_op_max_order,
            sv_mode,
            is_threshold,
        )

    # recombine everything
    return np.real(ker * integrand)


@nb.njit(cache=True)
def quad_ker_qcd(
    ker_base,
    order,
    mode0,
    mode1,
    method,
    as1,
    as0,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
    is_polarized,
    is_time_like,
    n3lo_ad_variation,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    quad_ker : float
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    as1 : float
        target coupling value
    as0 : float
        initial coupling value
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an itermediate threshold operator?
    n3lo_ad_variation : tuple
        |N3LO| anomalous dimension variation ``(gg_var, gq_var, qg_var, qq_var)``

    Returns
    -------
    float
        evaluated integration kernel
    """
    # compute the actual evolution kernel for pure QCD
    if ker_base.is_singlet:
        if is_polarized:
            if is_time_like:
                raise NotImplementedError("Polarized, time-like is not implemented")
            else:
                gamma_singlet = ad_ps.gamma_singlet(order, ker_base.n, nf)
        else:
            if is_time_like:
                gamma_singlet = ad_ut.gamma_singlet(order, ker_base.n, nf)
            else:
                gamma_singlet = ad_us.gamma_singlet(
                    order, ker_base.n, nf, n3lo_ad_variation
                )
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_singlet = sv.exponentiated.gamma_variation(
                gamma_singlet, order, nf, L
            )
        ker = s.dispatcher(
            order,
            method,
            gamma_singlet,
            as1,
            as0,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv.expanded.singlet_variation(gamma_singlet, as1, order, nf, L, dim=2)
            ) @ np.ascontiguousarray(ker)
        ker = select_singlet_element(ker, mode0, mode1)
    else:
        if is_polarized:
            if is_time_like:
                raise NotImplementedError("Polarized, time-like is not implemented")
            else:
                gamma_ns = ad_ps.gamma_ns(order, mode0, ker_base.n, nf)
        else:
            if is_time_like:
                gamma_ns = ad_ut.gamma_ns(order, mode0, ker_base.n, nf)
            else:
                gamma_ns = ad_us.gamma_ns(order, mode0, ker_base.n, nf)
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv.exponentiated.gamma_variation(gamma_ns, order, nf, L)
        ker = ns.dispatcher(
            order,
            method,
            gamma_ns,
            as1,
            as0,
            nf,
            ev_op_iterations,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = sv.expanded.non_singlet_variation(gamma_ns, as1, order, nf, L) * ker
    return ker


@nb.njit(cache=True)
def quad_ker_qed(
    ker_base,
    order,
    mode0,
    mode1,
    method,
    as_list,
    mu2_from,
    mu2_to,
    a_half,
    alphaem_running,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    ker_base : QuadKerBase
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    as1 : float
        target coupling value
    as0 : float
        initial coupling value
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2
    aem_list : list
        list of electromagnetic coupling values
    alphaem_running : bool
        whether alphaem is running or not
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an itermediate threshold operator?

    Returns
    -------
    float
        evaluated integration kernel
    """
    # compute the actual evolution kernel for QEDxQCD
    if ker_base.is_QEDsinglet:
        gamma_s = ad_us.gamma_singlet_qed(order, ker_base.n, nf)
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_s = sv.exponentiated.gamma_variation_qed(
                gamma_s, order, nf, L, alphaem_running
            )
        ker = qed_s.dispatcher(
            order,
            method,
            gamma_s,
            as_list,
            a_half,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        # TODO : in this way a_half[-1][1] is the aem value computed in
        # the middle point of the last step. Instead we want aem computed in mu2_final.
        # However the distance between the two is very small and affects only the running aem
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv.expanded.singlet_variation_qed(
                    gamma_s, as_list[-1], a_half[-1][1], alphaem_running, order, nf, L
                )
            ) @ np.ascontiguousarray(ker)
        ker = select_QEDsinglet_element(ker, mode0, mode1)
    elif ker_base.is_QEDvalence:
        gamma_v = ad_us.gamma_valence_qed(order, ker_base.n, nf)
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_v = sv.exponentiated.gamma_variation_qed(
                gamma_v, order, nf, L, alphaem_running
            )
        ker = qed_v.dispatcher(
            order,
            method,
            gamma_v,
            as_list,
            a_half,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv.expanded.valence_variation_qed(
                    gamma_v, as_list[-1], a_half[-1][1], alphaem_running, order, nf, L
                )
            ) @ np.ascontiguousarray(ker)
        ker = select_QEDvalence_element(ker, mode0, mode1)
    else:
        gamma_ns = ad_us.gamma_ns_qed(order, mode0, ker_base.n, nf)
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv.exponentiated.gamma_variation_qed(
                gamma_ns, order, nf, L, alphaem_running
            )
        ker = qed_ns.dispatcher(
            order,
            method,
            gamma_ns,
            as_list,
            a_half[:, 1],
            alphaem_running,
            nf,
            ev_op_iterations,
            mu2_from,
            mu2_to,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = (
                sv.expanded.non_singlet_variation_qed(
                    gamma_ns, as_list[-1], a_half[-1][1], alphaem_running, order, nf, L
                )
                * ker
            )
    return ker


class Operator(sv.ModeMixin):
    """Internal representation of a single EKO.

    The actual matrices are computed upon calling :meth:`compute`.

    Parameters
    ----------
    config : dict
        configuration
    managers : dict
        managers
    nf : int
        number of active flavors
    q2_from : float
        evolution source
    q2_to : float
        evolution target
    mellin_cut : float
        cut to the upper limit in the mellin inversion
    is_threshold : bool
        is this an itermediate threshold operator?
    """

    log_label = "Evolution"
    # complete list of possible evolution operators labels
    full_labels = br.full_labels
    full_labels_qed = br.full_unified_labels

    def __init__(
        self, config, managers, segment: Segment, mellin_cut=5e-2, is_threshold=False
    ):
        self.config = config
        self.managers = managers
        self.nf = segment.nf
        self.q2_from = segment.origin
        self.q2_to = segment.target
        # TODO make 'cut' external parameter?
        self._mellin_cut = mellin_cut
        self.is_threshold = is_threshold
        self.op_members = {}
        self.order = tuple(config["order"])
        self.alphaem_running = self.managers["couplings"].alphaem_running
        if self.log_label == "Evolution":
            self.a = self.compute_a()
            self.compute_aem_list()

    @property
    def n_pools(self):
        """Return number of parallel cores."""
        n_pools = self.config["n_integration_cores"]
        if n_pools > 0:
            return n_pools
        # so we subtract from the maximum number
        return max(os.cpu_count() + n_pools, 1)

    @property
    def xif2(self):
        r"""Return scale variation factor :math:`(\mu_F/\mu_R)^2`."""
        return self.config["xif2"]

    @property
    def int_disp(self):
        """Return the interpolation dispatcher."""
        return self.managers["interpol_dispatcher"]

    @property
    def grid_size(self):
        """Return the grid size."""
        return self.int_disp.xgrid.size

    @property
    def mu2(self):
        """Return the arguments to the :math:`a_s` function."""
        mu0 = self.q2_from * (
            self.xif2 if self.sv_mode == sv.Modes.exponentiated else 1.0
        )
        mu1 = self.q2_to * (
            self.xif2
            if (not self.is_threshold and self.sv_mode == sv.Modes.expanded)
            or self.sv_mode == sv.Modes.exponentiated
            else 1.0
        )
        return (mu0, mu1)

    def compute_a(self):
        """Return the computed values for :math:`a_s` and :math:`a_{em}`."""
        coupling = self.managers["couplings"]
        a0 = coupling.a(
            self.mu2[0],
            nf_to=self.nf,
        )
        a1 = coupling.a(
            self.mu2[1],
            nf_to=self.nf,
        )
        return (a0, a1)

    @property
    def a_s(self):
        """Return the computed values for :math:`a_s`."""
        return (self.a[0][0], self.a[1][0])

    @property
    def a_em(self):
        """Return the computed values for :math:`a_{em}`."""
        return (self.a[0][1], self.a[1][1])

    def compute_aem_list(self):
        """
        Return the list of the couplings for the different values of :math:`a_s`.

        This functions is needed in order to compute the values of :math:`a_s`
        and :math:`a_em` in the middle point of the :math:`mu^2` interval, and
        the values of :math:`a_s` at the borders of every intervals.
        This is needed in the running_alphaem solution.

        """
        ev_op_iterations = self.config["ev_op_iterations"]
        if self.order[1] == 0:
            self.as_list = np.array([self.a_s[0], self.a_s[1]])
            self.a_half_list = np.zeros((ev_op_iterations, 2))
        else:
            as0 = self.a_s[0]
            as1 = self.a_s[1]
            aem0 = self.a_em[0]
            aem1 = self.a_em[1]
            q2ref = self.managers["couplings"].mu2_ref
            delta_from = abs(self.q2_from - q2ref)
            delta_to = abs(self.q2_to - q2ref)
            # I compute the values in aem_list starting from the mu2
            # that is closer to mu_ref.
            if delta_from > delta_to:
                a_start = np.array([as1, aem1])
                mu2_start = self.q2_to
            else:
                a_start = np.array([as0, aem0])
                mu2_start = self.q2_from
            couplings = self.managers["couplings"]
            mu2_steps = utils.geomspace(self.q2_from, self.q2_to, 1 + ev_op_iterations)
            mu2_l = mu2_steps[0]
            self.as_list = np.array(
                [
                    couplings.compute(
                        a_ref=a_start, nf=self.nf, scale_from=mu2_start, scale_to=mu2
                    )[0]
                    for mu2 in mu2_steps
                ]
            )
            a_half = np.zeros((ev_op_iterations, 2))
            for step, mu2_h in enumerate(mu2_steps[1:]):
                mu2_half = (mu2_h + mu2_l) / 2.0
                a_s, aem = couplings.compute(
                    a_ref=a_start, nf=self.nf, scale_from=mu2_start, scale_to=mu2_half
                )
                a_half[step] = [a_s, aem]
                mu2_l = mu2_h
            self.a_half_list = a_half

    @property
    def labels(self):
        """Compute necessary sector labels to compute.

        Returns
        -------
        list(str)
            sector labels
        """
        labels = []
        # the NS sector is dynamic
        if self.order[1] == 0:
            if self.config["debug_skip_non_singlet"]:
                logger.warning("%s: skipping non-singlet sector", self.log_label)
            else:
                # add + as default
                labels.append(br.non_singlet_labels[1])
                if self.order[0] >= 2:  # - becomes different starting from NLO
                    labels.append(br.non_singlet_labels[0])
                if self.order[0] >= 3:  # v also becomes different starting from NNLO
                    labels.append(br.non_singlet_labels[2])
            # singlet sector is fixed
            if self.config["debug_skip_singlet"]:
                logger.warning("%s: skipping singlet sector", self.log_label)
            else:
                labels.extend(br.singlet_labels)
        else:
            if self.config["debug_skip_non_singlet"]:
                logger.warning("%s: skipping non-singlet sector", self.log_label)
            else:
                # add +u and +d as default
                labels.append(br.non_singlet_unified_labels[0])
                labels.append(br.non_singlet_unified_labels[2])
                # -u and -d become different starting from O(as1aem1) or O(aem2)
                if self.order[1] >= 2 or self.order[0] >= 1:
                    labels.append(br.non_singlet_unified_labels[1])
                    labels.append(br.non_singlet_unified_labels[3])
                labels.extend(br.valence_unified_labels)
            if self.config["debug_skip_singlet"]:
                logger.warning("%s: skipping singlet sector", self.log_label)
            else:
                labels.extend(br.singlet_unified_labels)
        return labels

    def quad_ker(self, label, logx, areas):
        """Return partially initialized integrand function.

        Parameters
        ----------
        label: tuple
            operator element pids
        logx: float
            Mellin inversion point
        areas : tuple
            basis function configuration

        Returns
        -------
        functools.partial
            partially initialized integration kernel

        """
        return functools.partial(
            quad_ker,
            order=self.order,
            mode0=label[0],
            mode1=label[1],
            method=self.config["method"],
            is_log=self.int_disp.log,
            logx=logx,
            areas=areas,
            as_list=self.as_list,
            mu2_from=self.q2_from,
            mu2_to=self.q2_to,
            a_half=self.a_half_list,
            alphaem_running=self.alphaem_running,
            nf=self.nf,
            L=np.log(self.xif2),
            ev_op_iterations=self.config["ev_op_iterations"],
            ev_op_max_order=tuple(self.config["ev_op_max_order"]),
            sv_mode=self.sv_mode,
            is_threshold=self.is_threshold,
            n3lo_ad_variation=self.config["n3lo_ad_variation"],
            is_polarized=self.config["polarized"],
            is_time_like=self.config["time_like"],
        )

    def initialize_op_members(self):
        """Init all operators with the identity or zeros."""
        eye = OpMember(
            np.eye(self.grid_size), np.zeros((self.grid_size, self.grid_size))
        )
        zero = OpMember(*[np.zeros((self.grid_size, self.grid_size))] * 2)
        if self.order[1] == 0:
            full_labels = self.full_labels
            non_singlet_labels = br.non_singlet_labels
        else:
            full_labels = self.full_labels_qed
            non_singlet_labels = br.non_singlet_unified_labels
        for n in full_labels:
            if n in self.labels:
                # non-singlet evolution and diagonal op are identities
                if n in non_singlet_labels or n[0] == n[1]:
                    self.op_members[n] = eye.copy()
                else:
                    self.op_members[n] = zero.copy()
            else:
                self.op_members[n] = zero.copy()

    def run_op_integration(
        self,
        log_grid,
    ):
        """Run the integration for each grid point.

        Parameters
        ----------
        log_grid : tuple(k, logx)
            log grid point with relative index

        Returns
        -------
        list
            computed operators at the give grid point
        """
        column = []
        k, logx = log_grid
        start_time = time.perf_counter()
        # iterate basis functions
        for l, bf in enumerate(self.int_disp):
            if k == l and l == self.grid_size - 1:
                continue
            temp_dict = {}
            # iterate sectors
            for label in self.labels:
                res = integrate.quad(
                    self.quad_ker(
                        label=label, logx=logx, areas=bf.areas_representation
                    ),
                    0.5,
                    1.0 - self._mellin_cut,
                    epsabs=1e-12,
                    epsrel=1e-5,
                    limit=100,
                    full_output=1,
                )
                temp_dict[label] = res[:2]
            column.append(temp_dict)
        logger.info(
            "%s: computing operators - %u/%u took: %6f s",
            self.log_label,
            k + 1,
            self.grid_size,
            (time.perf_counter() - start_time),
        )
        return column

    def compute(self):
        """Compute the actual operators (i.e. run the integrations)."""
        self.initialize_op_members()

        # skip computation?
        if np.isclose(self.q2_from, self.q2_to):
            # unless we have to do some scale variation
            # TODO remove if K is factored out of here
            if not (
                self.sv_mode == sv.Modes.expanded
                and not np.isclose(self.xif2, 1.0)
                and not self.is_threshold
            ):
                logger.info(
                    "%s: skipping unity operator at %e", self.log_label, self.q2_from
                )
                self.copy_ns_ops()
                return

        logger.info(
            "%s: computing operators %e -> %e, nf=%d",
            self.log_label,
            self.q2_from,
            self.q2_to,
            self.nf,
        )
        logger.info(
            "%s: µ_R^2 distance: %e -> %e",
            self.log_label,
            self.mu2[0],
            self.mu2[1],
        )
        if self.sv_mode != sv.Modes.unvaried:
            logger.info(
                "Scale Variation: (µ_F/µ_R)^2 = %e, mode: %s",
                self.xif2,
                self.sv_mode.name,
            )
        logger.info(
            "%s: a_s distance: %e -> %e", self.log_label, self.a_s[0], self.a_s[1]
        )
        if self.order[1] > 0:
            logger.info(
                "%s: a_em distance: %e -> %e",
                self.log_label,
                self.a_em[0],
                self.a_em[1],
            )
        logger.info(
            "%s: order: (%d, %d), solution strategy: %s",
            self.log_label,
            self.order[0],
            self.order[1],
            self.config["method"],
        )

        self.integrate()
        # copy non-singlet kernels, if necessary
        self.copy_ns_ops()

    def integrate(
        self,
    ):
        """Run the integration."""
        tot_start_time = time.perf_counter()

        # run integration in parallel for each grid point
        # or avoid opening a single pool
        args = (self.run_op_integration, enumerate(np.log(self.int_disp.xgrid.raw)))
        if self.n_pools == 1:
            res = map(*args)
        else:
            with Pool(self.n_pools) as pool:
                res = pool.map(*args)

        # collect results
        for k, row in enumerate(res):
            for l, entry in enumerate(row):
                for label, (val, err) in entry.items():
                    self.op_members[label].value[k][l] = val
                    self.op_members[label].error[k][l] = err

        # closing comment
        logger.info(
            "%s: Total time %f s",
            self.log_label,
            time.perf_counter() - tot_start_time,
        )

    def copy_ns_ops(self):
        """Copy non-singlet kernels, if necessary."""
        if self.order[1] == 0:
            if self.order[0] == 1:  # in LO +=-=v
                for label in ["nsV", "ns-"]:
                    self.op_members[
                        (br.non_singlet_pids_map[label], 0)
                    ].value = self.op_members[
                        (br.non_singlet_pids_map["ns+"], 0)
                    ].value.copy()
                    self.op_members[
                        (br.non_singlet_pids_map[label], 0)
                    ].error = self.op_members[
                        (br.non_singlet_pids_map["ns+"], 0)
                    ].error.copy()
            elif self.order[0] == 2:  # in NLO -=v
                self.op_members[
                    (br.non_singlet_pids_map["nsV"], 0)
                ].value = self.op_members[
                    (br.non_singlet_pids_map["ns-"], 0)
                ].value.copy()
                self.op_members[
                    (br.non_singlet_pids_map["nsV"], 0)
                ].error = self.op_members[
                    (br.non_singlet_pids_map["ns-"], 0)
                ].error.copy()
        # at O(as0aem1) u-=u+, d-=d+
        # starting from O(as1aem1) P+ != P-
        # However the solution with pure QED is not implemented in EKO
        # so the ns anomalous dimensions are always different
        # elif self.order[1] == 1 and self.order[0] == 0:
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-u"], 0)
        #     ].value = self.op_members[(br.non_singlet_pids_map["ns+u"], 0)].value.copy()
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-u"], 0)
        #     ].error = self.op_members[(br.non_singlet_pids_map["ns+u"], 0)].error.copy()
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-d"], 0)
        #     ].value = self.op_members[(br.non_singlet_pids_map["ns+d"], 0)].value.copy()
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-d"], 0)
        #     ].error = self.op_members[(br.non_singlet_pids_map["ns+d"], 0)].error.copy()



"""IO generic exceptions."""
import os


class OutputError(Exception):
    """Generic Output Error."""


class OutputExistsError(FileExistsError, OutputError):
    """Output file already exists."""


class OutputNotTar(ValueError, OutputError):
    """Specified file is not a .tar archive."""


class OperatorLoadingError(ValueError, OutputError):
    """Issue encountered while loading an operator."""


class OperatorLocationError(ValueError, OutputError):
    """Path supposed to store an operator in wrong location."""

    def __init__(self, path: os.PathLike):
        self.path = path
        super().__init__(f"Path '{path}' not in operators folder")



"""Implementation of :cite:`Giele:2002hx` and  :cite:`Dittmar:2005ed` (NNLO and polarized)."""
import pathlib

import numpy as np
import yaml
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from eko import basis_rotation as br

from ...plots import plot_dist

here = pathlib.Path(__file__).parents[0]

# xgrid
toy_xgrid = np.array([1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 0.1, 0.3, 0.5, 0.7, 0.9])

raw_label_list = ["u_v", "d_v", "L_m", "L_p", "s_p", "c_p", "b_p", "g"]

# rot_func_list = [toy_V0, toy_V30, toy_T30, toy_T80, toy_S0, toy_S0, toy_S0, toy_g0]

# my/exact initial grid
# LHA_init_grid = []
# for f in [toy_uv0, toy_dv0, toy_Lm0, toy_Lp0, toy_sp0, toy_cp0, toy_bp0, toy_g0]:
#    LHA_init_grid.append(f(toy_xgrid))
LHA_init_grid = np.array([])


# L+ = 2(ub + db) = u+ - u- + d+ - d-
# L- = db - ub = ((d+ - d-) - (u+-u-))/2
# In the NNLO paper :cite:`Dittmar:2005ed` L_+ definition is different:
# L_+ = 2 L_+_NNLO
LHA_rotate_to_flavor = np.array(
    [
        # u_v, d_v, L_-, L_+, s_+, c_+, b_+,   g
        [0, 0, 0, 0, 0, 0, 0, 0],  # ph
        [0, 0, 0, 0, 0, 0, 0, 0],  # tbar
        [0, 0, 0, 0, 0, 0, 1 / 2, 0],  # bbar
        [0, 0, 0, 0, 0, 1 / 2, 0, 0],  # cbar
        [0, 0, 0, 0, 1 / 2, 0, 0, 0],  # sbar
        [0, 0, -1 / 2, 1 / 4, 0, 0, 0, 0],  # ubar
        [0, 0, 1 / 2, 1 / 4, 0, 0, 0, 0],  # dbar
        [0, 0, 0, 0, 0, 0, 0, 1],  # g
        [0, 1, 1 / 2, 1 / 4, 0, 0, 0, 0],  # d
        [1, 0, -1 / 2, 1 / 4, 0, 0, 0, 0],  # u
        [0, 0, 0, 0, 1 / 2, 0, 0, 0],  # s
        [0, 0, 0, 0, 0, 1 / 2, 0, 0],  # c
        [0, 0, 0, 0, 0, 0, 1 / 2, 0],  # b
        [0, 0, 0, 0, 0, 0, 0, 0],  # t
    ]
)


# rotate basis
def rotate_data(raw, is_ffns_nnlo=False, rotate_to_evolution_basis=False):
    """Rotate data in flavor space.

    Rotate either to flavor basis  or evolution basis from the LHA basis, which is yet an other basis.

    Parameters
    ----------
    raw : dict
        data
    is_ffns_nnlo : bool
        special table for NNLO FFNS
    rotate_to_evolution_basis : bool
        to evolution basis?

    Returns
    -------
    dict
        rotated data

    """
    inp = []
    label_list = raw_label_list
    to_flavor = LHA_rotate_to_flavor
    to_evolution = np.copy(br.rotate_flavor_to_evolution)
    if is_ffns_nnlo:
        # add s_v and delete b_p to label_list
        label_list = np.insert(label_list, 4, "s_v")
        label_list = np.delete(label_list, -2)

        # change the rotation matrix
        b_line = [0, 0, 0, 0, 0, 0, 1 / 2, 0]
        # b to zeros
        to_flavor = np.where(to_flavor == b_line, np.zeros(8), to_flavor)
        # c_p to b_p position
        to_flavor[3, :] = b_line
        to_flavor[-3, :] = b_line
        # s
        to_flavor[4, :] = [0, 0, 0, 0, -1 / 2, 1 / 2, 0, 0]
        to_flavor[-4, :] = [0, 0, 0, 0, 1 / 2, 1 / 2, 0, 0]

        # s_v = c_v count twice
        to_evolution[3, 4] = -2
        to_evolution[3, -4] = 2
    for l in label_list:
        inp.append(raw[l])
    inp = np.array(inp)
    flav_pdfs = np.dot(to_flavor, inp)

    # additional rotation to evolution basis if necessary
    if rotate_to_evolution_basis:
        evol_pdfs = np.matmul(to_evolution, flav_pdfs)
        return dict(zip(br.evol_basis, evol_pdfs))
    return dict(zip(br.flavor_basis_pids, flav_pdfs))


def compute_LHA_data(theory, operators, rotate_to_evolution_basis=False):
    """Implement LHA benchmark :cite:`Giele:2002hx`.

    Parameters
    ----------
    theory : dict
        theory card
    operators : dict
        operators card
    rotate_to_evolution_basis : bool
        rotate to evolution basis

    Returns
    -------
    dict
        output containing: target_xgrid, values

    """
    polarized = operators["polarized"]
    mu2grid = np.power(operators["mugrid"], 2.0)
    if not np.allclose(mu2grid, [1e4]):
        raise ValueError("mu2grid has to be [1e4]")
    order = theory["PTO"]
    # select which data
    if polarized and order <= 1:
        yaml_file = "LHA_polarized.yaml"
    elif polarized and order > 1:
        raise ValueError("LHA tables beyond NLO do not exist for polarized Case")
    else:
        yaml_file = "LHA.yaml"
    # load data
    with open(here / yaml_file, encoding="utf-8") as o:
        data = yaml.safe_load(o)
    fns = theory["FNS"]
    xif2 = (theory["XIF"]) ** 2
    if order == 0 and xif2 != 1.0:
        raise ValueError("LO LHA tables with scale variations are not available")
    table = None
    part = None
    is_ffns_nnlo = False

    # Switching at the intermediate point.
    if xif2 > np.sqrt(2):
        part = 2
    elif xif2 < np.sqrt(1.0 / 2.0):
        part = 3
    else:
        part = 1
    if fns == "FFNS":
        if order == 0:
            part = 2
            table = 16 if polarized else 2
        elif order == 1:
            table = 17 if polarized else 3
        elif order == 2:
            is_ffns_nnlo = True
            table = 14
    elif fns == "ZM-VFNS":
        if order == 0:
            part = 3
            table = 16 if polarized else 2
        elif order == 1:
            table = 18 if polarized else 4
        elif order == 2:
            table = 15
    else:
        raise ValueError(f"unknown FNS {fns} or order {order}")
    ref_values = rotate_data(
        data[f"table{table}"][f"part{part}"], is_ffns_nnlo, rotate_to_evolution_basis
    )
    ref = {
        "target_xgrid": toy_xgrid,
        "values": {1e4: ref_values},
    }

    return ref


def save_initial_scale_plots_to_pdf(path, is_pol):
    r"""Plot all PDFs at the initial scale.

    The reference values are given in Table 2 part 1 :cite:`Giele:2002hx`
    or Table 16 part 1 (polarized) :cite:`Dittmar:2005ed` .

    This exercise was useful in order to detect the missing 2 in the definition of
    :math:`L_+ = 2(\\bar u + \\bar d)`

    Parameters
    ----------
    path : str
        output path
    is_pol : bool
        polarized pdf data

    """
    # load data
    if not is_pol:
        yaml_file = "LHA.yaml"
        table = "table2"
    else:
        yaml_file = "LHA_polarized.yaml"
        table = "table16"
    with open(here / yaml_file, encoding="utf-8") as o:
        data = yaml.safe_load(o)
    LHA_init_grid_ref = data[table]["part1"]
    with PdfPages(path) as pp:
        # iterate all raw labels
        for j, label in enumerate(raw_label_list):
            # skip trivial plots
            if label in ["c_p", "b_p"]:
                continue
            me = LHA_init_grid[j]
            ref = LHA_init_grid_ref[label]
            fig = plot_dist(
                toy_xgrid,
                toy_xgrid * me,
                np.zeros(len(me)),
                ref,
                title=f"x{label}(x,µ_F^2 = 2 GeV^2)",
            )
            pp.savefig()
            plt.close(fig)
    print(f"Initial scale pdf plots written to {path}")



"""Fully managed runner.

This is an automated runner, mainly suggested for small EKOs computations.

The primitives used here to compute the various pieces are part of the public
interface, and should be directly used to manage a more complex run for a
considebaly large operator.

Thus, parallelization and multi-node execution is possible using EKO primitives,
but not automatically performed.

"""
from pathlib import Path

from ..io.items import Evolution, Matching, Target
from ..io.runcards import OperatorCard, TheoryCard
from ..io.struct import EKO
from . import commons, operators, parts, recipes


def solve(theory: TheoryCard, operator: OperatorCard, path: Path):
    """Solve DGLAP equations in terms of evolution kernel operators (EKO)."""
    theory.heavy.intrinsic_flavors = [4, 5, 6]

    with EKO.create(path) as builder:
        eko = builder.load_cards(theory, operator).build()  # pylint: disable=E1101

        atlas = commons.atlas(eko.theory_card, eko.operator_card)

        recs = recipes.create(eko.operator_card.evolgrid, atlas)
        eko.load_recipes(recs)

        for recipe in eko.recipes:
            assert isinstance(recipe, Evolution)
            eko.parts[recipe] = parts.evolve(eko, recipe)
            # flush the memory
            del eko.parts[recipe]
        for recipe in eko.recipes_matching:
            assert isinstance(recipe, Matching)
            eko.parts_matching[recipe] = parts.match(eko, recipe)
            # flush the memory
            del eko.parts_matching[recipe]

        for ep in operator.evolgrid:
            headers = recipes.elements(ep, atlas)
            parts_ = operators.retrieve(headers, eko.parts, eko.parts_matching)
            target = Target.from_ep(ep)
            eko.operators[target] = operators.join(parts_)
            # flush the memory
            del eko.parts
            del eko.parts_matching
            del eko.operators[target]



"""Physical and internal objects.

This subpackage contains the definitions of many objects, with the purpose of
using them as type hints, but not only (at least for some of them).

Few related functions are also defined here.

"""



r"""Scale variation operator for the expanded scheme (``ModSV=expanded``).

The expressions can be obtained using Eqs. (3.33) and (3.38) of :cite:`AbdulKhalek:2019ihb`.
Be aware that corresponding the signs of the ingredients there are a number of differences.
However, the ultimate sign can be obtained by comparing to the exponentiated scheme in the
trunctated solution.
"""

import numba as nb
import numpy as np

from .. import beta


@nb.njit(cache=True)
def variation_as1(gamma, L):
    r"""Compute the |NLO| anomalous dimension variation.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    complex
        variation at |NLO|
    """
    return L * gamma[0]


@nb.njit(cache=True)
def variation_as2(gamma, L, beta0, g0e2):
    r"""Compute the |NNLO| anomalous dimension variation.

    These kernels are meant to be used with alpha_s evaluated at the
    factorization scale. If one expresses everything in terms of
    alpha_s evaluated at the process scale, the sign of beta*gamma0 flips.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    L : float
        logarithmic ratio of factorization and renormalization scale
    beta0: float
        :math:`\beta_0`
    g0e2: complex or numpy.ndarray
        :math:`\left(\gamma^{(0)}\right)^2`

    Returns
    -------
    complex
        variation at |NNLO|
    """
    return gamma[1] * L + 1.0 / 2.0 * (beta0 * gamma[0] + g0e2) * L**2


@nb.njit(cache=True)
def variation_as3(gamma, L, beta0, beta1, g0e2, g0e3, g1g0, g0g1):
    r"""Compute the |N3LO| anomalous dimension variation.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    L : float
        logarithmic ratio of factorization and renormalization scale
    beta0: float
        :math:`\beta_0`
    beta0: float
        :math:`\beta_1`
    g0e2: complex or numpy.ndarray
        :math:`\left(\gamma^{(0)}\right)^2`
    g0e3: complex or numpy.ndarray
        :math:`\left(\gamma^{(0)}\right)^3`
    g1g0: complex or numpy.ndarray
        :math:`\gamma^{(1)} \gamma^{(0)}`
    g0g1: complex or numpy.ndarray
        :math:`\gamma^{(0)} \gamma^{(1)}`

    Returns
    -------
    complex
        variation at |N3LO|
    """
    return (
        gamma[2] * L
        + (1.0 / 2.0)
        * (beta1 * gamma[0] + 2.0 * beta0 * gamma[1] + g1g0 + g0g1)
        * L**2
        + (1.0 / 6.0)
        * (2.0 * beta0**2 * gamma[0] + 3.0 * beta0 * g0e2 + g0e3)
        * L**3
    )


@nb.njit(cache=True)
def non_singlet_variation(gamma, a_s, order, nf, L):
    """Non-singlet scale variation dispatcher.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    a_s :  float
        target coupling value
    order : int
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    complex
        scale variation kernel
    """
    sv_ker = 1.0
    if order[0] >= 2:
        sv_ker += a_s * variation_as1(gamma, L)
    if order[0] >= 3:
        beta0 = beta.beta_qcd_as2(nf)
        sv_ker += a_s**2 * variation_as2(gamma, L, beta0, gamma[0] ** 2)
    if order[0] >= 4:
        beta1 = beta.beta_qcd((3, 0), nf)
        g0g1 = gamma[0] * gamma[1]
        sv_ker += a_s**3 * variation_as3(
            gamma, L, beta0, beta1, gamma[0] ** 2, gamma[0] ** 3, g0g1, g0g1
        )
    return sv_ker


@nb.njit(cache=True)
def singlet_variation(gamma, a_s, order, nf, L, dim):
    """Singlet scale variation dispatcher.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    a_s :  float
        target coupling value
    order : int
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    numpy.ndarray
        scale variation kernel
    """
    sv_ker = np.eye(dim, dtype=np.complex_)
    gamma = np.ascontiguousarray(gamma)
    if order[0] >= 2:
        sv_ker += a_s * variation_as1(gamma, L)
    if order[0] >= 3:
        beta0 = beta.beta_qcd_as2(nf)
        gamma0e2 = gamma[0] @ gamma[0]
        sv_ker += a_s**2 * variation_as2(gamma, L, beta0, gamma0e2)
    if order[0] >= 4:
        beta1 = beta.beta_qcd((3, 0), nf)
        gamma0e3 = gamma0e2 @ gamma[0]
        # here the product is not commutative
        g1g0 = gamma[1] @ gamma[0]
        g0g1 = gamma[0] @ gamma[1]
        sv_ker += a_s**3 * variation_as3(
            gamma, L, beta0, beta1, gamma0e2, gamma0e3, g1g0, g0g1
        )
    return sv_ker


@nb.njit(cache=True)
def non_singlet_variation_qed(gamma, a_s, a_em, alphaem_running, order, nf, L):
    """Non-singlet scale variation dispatcher.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    a_s :  float
        target coupling value
    order : int
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    complex
        scale variation kernel
    """
    sv_ker = non_singlet_variation(gamma[1:, 0], a_s, order, nf, L)
    if alphaem_running:
        if order[1] >= 2:
            sv_ker += a_em * variation_as1(gamma[0, 1:], L)
    return sv_ker


@nb.njit(cache=True)
def singlet_variation_qed(gamma, a_s, a_em, alphaem_running, order, nf, L):
    """Singlet scale variation dispatcher.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    a_s :  float
        target coupling value
    order : int
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    numpy.ndarray
        scale variation kernel
    """
    sv_ker = singlet_variation(gamma[1:, 0], a_s, order, nf, L, 4)
    if alphaem_running:
        if order[1] >= 2:
            sv_ker += a_em * variation_as1(gamma[0, 1:], L)
    return sv_ker


@nb.njit(cache=True)
def valence_variation_qed(gamma, a_s, a_em, alphaem_running, order, nf, L):
    """Singlet scale variation dispatcher.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    a_s :  float
        target coupling value
    order : int
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    numpy.ndarray
        scale variation kernel
    """
    sv_ker = singlet_variation(gamma[1:, 0], a_s, order, nf, L, 2)
    if alphaem_running:
        if order[1] >= 2:
            sv_ker += a_em * variation_as1(gamma[0, 1:], L)
    return sv_ker



"""The :math:`O(a_{em}^1)` Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko import constants

from . import as1


@nb.njit(cache=True)
def gamma_phq(N):
    r"""Compute the leading-order photon-quark anomalous dimension.

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    complex
        Leading-order photon-quark anomalous dimension :math:`\\gamma_{\\gamma q}^{(0,1)}(N)`

    """
    return as1.gamma_gq(N) / constants.CF


@nb.njit(cache=True)
def gamma_qph(N, nf):
    r"""Compute the leading-order quark-photon anomalous dimension.

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.
    But adding the :math:`N_C` and the :math:`2n_f` factors from :math:`\\theta` inside the
    definition of :math:`\\gamma_{q \\gamma}^{(0)}(N)`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        Leading-order quark-photon anomalous dimension :math:`\\gamma_{q \\gamma}^{(0,1)}(N)`

    """
    return as1.gamma_qg(N, nf) / constants.TR * constants.NC


@nb.njit(cache=True)
def gamma_phph(nf):
    r"""Compute the leading-order photon-photon anomalous dimension.

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.

    Parameters
    ----------
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        Leading-order phton-photon anomalous dimension :math:`\\gamma_{\\gamma \\gamma}^{(0,1)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return 4.0 / 3.0 * constants.NC * (nu * constants.eu2 + nd * constants.ed2)


@nb.njit(cache=True)
def gamma_ns(N, cache):
    r"""Compute the leading-order non-singlet QED anomalous dimension.

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        Leading-order non-singlet QED anomalous dimension :math:`\\gamma_{ns}^{(0,1)}(N)`

    """
    return as1.gamma_ns(N, cache) / constants.CF


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache):
    r"""Compute the QED leading-order singlet anomalous dimension matrix.

    .. math::
        \\gamma_S^{(0)} = \\left(\begin{array}{cc}
        0 & 0 & 0 & 0 \\
        0 & \\gamma_{\\gamma \\gamma}^{(0,1)} & \\langle e^2 \rangle \\gamma_{\\gamma q}^{(0,1)} & \nu_u e^2_- \\gamma_{\\gamma q}^{(0,1)}\\
        0 & \\langle e^2 \rangle\\gamma_{q \\gamma}^{(0,1)} & \\langle e^2 \rangle \\gamma_{ns}^{(0,1)} & \nu_u e^2_- \\gamma_{ns}^{(0,1)}\\
        0 & \nu_d e^2_- \\gamma_{q \\gamma}^{(0,1)} & \nu_d e^2_- \\gamma_{ns}^{(0,1)} & e^2_\\Delta \\gamma_{ns}^{(0,1)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(0)}(N)`

    """
    e2avg, vue2m, vde2m, e2delta = constants.charge_combinations(nf)
    gamma_ph_q = gamma_phq(N)
    gamma_q_ph = gamma_qph(N, nf)
    gamma_nonsinglet = gamma_ns(N, cache)
    gamma_S_01 = np.array(
        [
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [
                0.0 + 0.0j,
                gamma_phph(nf),
                e2avg * gamma_ph_q,
                vue2m * gamma_ph_q,
            ],
            [
                0.0 + 0.0j,
                e2avg * gamma_q_ph,
                e2avg * gamma_nonsinglet,
                vue2m * gamma_nonsinglet,
            ],
            [
                0.0 + 0.0j,
                vde2m * gamma_q_ph,
                vde2m * gamma_nonsinglet,
                e2delta * gamma_nonsinglet,
            ],
        ],
        np.complex_,
    )
    return gamma_S_01


@nb.njit(cache=True)
def gamma_valence(N, nf, cache):
    r"""Compute the QED leading-order valence anomalous dimension matrix.

    .. math::
        \\gamma_V^{(0,1)} = \\left(\begin{array}{cc}
        \\langle e^2 \rangle \\gamma_{ns}^{(0,1)} & \nu_u e^2_- \\gamma_{ns}^{(0,1)}\\
        \nu_d e^2_- \\gamma_{ns}^{(0,1)} & e^2_\\Delta \\gamma_{ns}^{(0,1)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(0)}(N)`

    """
    e2avg, vue2m, vde2m, e2delta = constants.charge_combinations(nf)
    gamma_V_01 = np.array(
        [
            [e2avg, vue2m],
            [vde2m, e2delta],
        ],
        np.complex_,
    )
    return gamma_V_01 * gamma_ns(N, cache)



"""The |N3LO| Altarelli-Parisi splitting kernels.

For further documentation see :doc:`N3LO anomalous dimensions <../../../theory/N3LO_ad>`

"""
import numba as nb
import numpy as np

from .ggg import gamma_gg
from .ggq import gamma_gq
from .gnsm import gamma_nsm
from .gnsp import gamma_nsp
from .gnsv import gamma_nsv
from .gps import gamma_ps
from .gqg import gamma_qg


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache, variation):
    r"""Compute the |N3LO| singlet anomalous dimension matrix.

      .. math::
          \gamma_S^{(3)} = \left(\begin{array}{cc}
          \gamma_{qq}^{(3)} & \gamma_{qg}^{(3)}\\
          \gamma_{gq}^{(3)} & \gamma_{gg}^{(3)}
          \end{array}\right)

    Parameters
    ----------
    N : complex
      Mellin moment
    nf : int
      Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache
    variation : tuple
        |N3LO| anomalous dimension variation ``(gg_var, gq_var, qg_var, qq_var)``

    Returns
    -------
    numpy.ndarray
        |N3LO| singlet anomalous dimension matrix
        :math:`\gamma_{S}^{(3)}(N)`

    """
    gamma_qq = gamma_nsp(N, nf, cache) + gamma_ps(N, nf, cache, variation[3])
    gamma_S_0 = np.array(
        [
            [gamma_qq, gamma_qg(N, nf, cache, variation[2])],
            [
                gamma_gq(N, nf, cache, variation[1]),
                gamma_gg(N, nf, cache, variation[0]),
            ],
        ],
        np.complex_,
    )
    return gamma_S_0



r"""|RGE| for the |MSbar| masses."""
from typing import List

import numba as nb
import numpy as np
from scipy import integrate, optimize

from .basis_rotation import quark_names
from .beta import b_qcd, beta_qcd
from .couplings import Couplings, invert_matching_coeffs
from .gamma import gamma
from .io.types import FlavorsNumber, Order
from .matchings import Atlas, flavor_shift, is_downward_path
from .quantities.couplings import CouplingEvolutionMethod, CouplingsInfo
from .quantities.heavy_quarks import HeavyQuarkMasses, QuarkMassRef, QuarkMassScheme


def ker_exact(a0, a1, order, nf):
    r"""Provide exact |MSbar| |RGE| kernel.

    Parameters
    ----------
    a0 : float
        strong coupling at the initial scale
    a1 : float
        strong coupling at the final scale
    oreder : tuple(int,int)
        perturbative order
    nf : int
        number of active flavours

    Returns
    -------
    float
        Exact |MSbar| kernel:

        .. math::
            k_{exact} = e^{-\int_{a_s(\mu_{h,0}^2)}^{a_s(\mu^2)}\gamma_m(a_s)/ \beta(a_s)da_s}

    """
    b_vec = [beta_qcd((2, 0), nf)]
    g_vec = [gamma(1, nf)]
    if order[0] >= 2:
        b_vec.append(beta_qcd((3, 0), nf))
        g_vec.append(gamma(2, nf))
    if order[0] >= 3:
        b_vec.append(beta_qcd((4, 0), nf))
        g_vec.append(gamma(3, nf))
    if order[0] >= 4:
        b_vec.append(beta_qcd((5, 0), nf))
        g_vec.append(gamma(4, nf))

    # quad ker
    def integrand(a, b_vec, g_vec):
        # minus sign goes away
        fgamma = np.sum([a**k * b for k, b in enumerate(g_vec)])
        fbeta = a * np.sum([a**k * b for k, b in enumerate(b_vec)])
        return fgamma / fbeta

    res = integrate.quad(
        integrand,
        a0,
        a1,
        args=(b_vec, g_vec),
        epsabs=1e-12,
        epsrel=1e-5,
        limit=100,
        full_output=1,
    )
    val, _ = res[:2]
    return np.exp(val)


@nb.njit(cache=True)
def ker_expanded(a0, a1, order, nf):
    r"""Provide expanded |MSbar| |RGE| kernel.

    Parameters
    ----------
    a0 : float
        strong coupling at the initial scale
    a1 : float
        strong coupling at the final scale
    order : tuple(int,int)
        perturbative order
    nf : int
        number of active flavours

    Returns
    -------
    float
        Expanded |MSbar| kernel:

        .. math::
            k_{expanded} &= \left (\frac{a_s(\mu^2)}{a_s(\mu_{h,0}^2)} \right )^{c_0}
            \frac{j_{exp}(a_s(\mu^2))}{j_{exp}(a_s(\mu_{h,0}^2))} \\
            j_{exp}(a_s) &= 1 + a_s \left [ c_1 - b_1 c_0 \right ] \\
            & + \frac{a_s^2}{2}
            \left [c_2 - c_1 b_1 - b_2 c_0 + b_1^2 c_0 + (c_1 - b_1 c_0)^2 \right ] \\
            & + \frac{a_s^3}{6} [ -2 b_3 c_0 - b_1^3 c_0 (1 + c_0) (2 + c_0) - 2 b_2 c_1 \\
            & - 3 b_2 c_0 c_1 + b_1^2 (2 + 3 c_0 (2 + c_0)) c_1 + c_1^3 + 3 c_1 c_2 \\
            & + b_1 (b_2 c_0 (4 + 3 c_0) - 3 (1 + c_0) c_1^2 - (2 + 3 c_0) c_2) + 2 c_3 ]

    """
    b0 = beta_qcd((2, 0), nf)
    c0 = gamma(1, nf) / b0
    ev_mass = np.power(a1 / a0, c0)
    num = 1.0
    den = 1.0
    if order[0] >= 2:
        b1 = b_qcd((3, 0), nf)
        c1 = gamma(2, nf) / b0
        u = c1 - b1 * c0
        num += a1 * u
        den += a0 * u
    if order[0] >= 3:
        b2 = b_qcd((4, 0), nf)
        c2 = gamma(3, nf) / b0
        u = (c2 - c1 * b1 - b2 * c0 + b1**2 * c0 + (c1 - b1 * c0) ** 2) / 2.0
        num += a1**2 * u
        den += a0**2 * u
    if order[0] >= 4:
        b3 = b_qcd((5, 0), nf)
        c3 = gamma(4, nf) / b0
        u = (
            1
            / 6
            * (
                -2 * b3 * c0
                - b1**3 * c0 * (1 + c0) * (2 + c0)
                - 2 * b2 * c1
                - 3 * b2 * c0 * c1
                + b1**2 * (2 + 3 * c0 * (2 + c0)) * c1
                + c1**3
                + 3 * c1 * c2
                + b1
                * (b2 * c0 * (4 + 3 * c0) - 3 * (1 + c0) * c1**2 - (2 + 3 * c0) * c2)
                + 2 * c3
            )
        )
        num += a1**3 * u
        den += a0**3 * u
    return ev_mass * num / den


def ker_dispatcher(q2_to, q2m_ref, strong_coupling, xif2, nf):
    r"""Select the |MSbar| kernel and compute the strong coupling values.

    Parameters
    ----------
    q2_to : float
        final scale
    q2m_ref : float
        initial scale
    strong_coupling : eko.strong_coupling.StrongCoupling
        Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
        any q
    xif2 : float
        factorization to renormalization scale ratio
    nf : int
        number of active flavours

    Returns
    -------
    ker :
        Expanded or exact |MSbar| kernel

    """
    a0 = strong_coupling.a(q2m_ref * xif2, nf)[0]
    a1 = strong_coupling.a(q2_to * xif2, nf)[0]
    method = strong_coupling.method
    order = strong_coupling.order
    if method == "expanded":
        return ker_expanded(a0, float(a1), order, nf)
    return ker_exact(a0, a1, order, nf)


def compute_matching_coeffs_up(nf: int):
    """Upward |MSbar| matching coefficients.

    Used at threshold when moving to a regime with *more* flavors
    :cite:`Liu:2015fxa`.

    Note :cite:`Liu:2015fxa` (eq 5.) reports the backward relation,
    multiplied by a factor of 4 (and 4^2 ...)

    Parameters
    ----------
    nf : int
        number of active flavors in the lower patch

    Returns
    -------
    dict
        downward matching coefficient matrix

    """
    matching_coeffs_up = np.zeros((4, 4))
    matching_coeffs_up[2, 0] = -89.0 / 27.0
    matching_coeffs_up[2, 1] = 20.0 / 9.0
    matching_coeffs_up[2, 2] = -4.0 / 3.0

    matching_coeffs_up[3, 0] = -118.248 - 1.58257 * nf
    matching_coeffs_up[3, 1] = 71.7887 + 7.85185 * nf
    matching_coeffs_up[3, 2] = -700.0 / 27.0
    matching_coeffs_up[3, 3] = -232.0 / 27.0 + 16.0 / 27.0 * nf

    return matching_coeffs_up


def compute_matching_coeffs_down(nf: int):
    """Downward |MSbar| matching coefficients.

    Used at threshold when moving to a regime with *less* flavors
    :cite:`Liu:2015fxa` :eqref:`5`.

    Parameters
    ----------
    nf : int
        number of active flavors in the lower patch

    Returns
    -------
    dict
        downward matching coefficient matrix

    """
    c_up = compute_matching_coeffs_up(nf)
    return invert_matching_coeffs(c_up)


def solve(m2_ref, q2m_ref, strong_coupling, nf_ref, xif2):
    r"""Compute the |MSbar| masses.

    Solves the equation :math:`m_{\overline{MS}}(m) = m` for a fixed number of
    nf.

    Parameters
    ----------
    m2_ref : float
        squared initial mass reference
    q2m_ref : float
        squared initial scale
    strong_coupling : eko.strong_coupling.StrongCoupling
        Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
        any q
    nf_ref : int
        number of active flavours at the scale q2m_ref, where the solution is searched
    xif2 : float
        :math:`\mu_F^2/\mu_R^2`

    Returns
    -------
    m2 : float
        :math:`m_{\overline{MS}}(\mu_2)^2`

    """

    def rge(m2, q2m_ref, strong_coupling, xif2, nf_ref):
        return (
            m2_ref * ker_dispatcher(m2, q2m_ref, strong_coupling, xif2, nf_ref) ** 2
            - m2
        )

    msbar_mass = optimize.fsolve(
        rge, q2m_ref, args=(q2m_ref, strong_coupling, xif2, nf_ref)
    )
    return float(msbar_mass)


def evolve(
    m2_ref,
    q2m_ref,
    strong_coupling,
    thresholds_ratios,
    xif2,
    q2_to,
    nf_ref=None,
    nf_to=None,
):
    r"""Perform the |MSbar| mass evolution up to given scale.

    It allows for different number of active flavors.

    Parameters
    ----------
    m2_ref : float
        squared initial mass reference
    q2m_ref : float
        squared initial scale
    strong_coupling : eko.strong_coupling.StrongCoupling
        Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
        any q
    xif2 : float
        :math:`\mu_F^2/\mu_R^2`
    q2_to : float
        scale at which the mass is computed
    nf_ref : int
        number of flavor active at the reference scale
    nf_to : int
        number of flavor active at the target scale

    Returns
    -------
    m2 : float
        :math:`m_{\overline{MS}}(\mu_2)^2`

    """
    matching_scales = np.array(strong_coupling.atlas.walls)[1:-1] * np.array(
        thresholds_ratios
    )
    atlas = Atlas(matching_scales.tolist(), (q2m_ref, nf_ref))
    path = atlas.path((q2_to, nf_to))
    is_downward = is_downward_path(path)
    shift = flavor_shift(is_downward)

    ev_mass = 1.0
    for k, seg in enumerate(path):
        # skip a very short segment, but keep the matching
        ker_evol = 1.0
        if not np.isclose(seg.origin, seg.target):
            ker_evol = (
                ker_dispatcher(seg.target, seg.origin, strong_coupling, xif2, seg.nf)
                ** 2
            )
        # apply matching condition
        if k < len(path) - 1:
            L = np.log(thresholds_ratios[seg.nf - shift])
            m_coeffs = (
                compute_matching_coeffs_down(seg.nf - 1)
                if is_downward
                else compute_matching_coeffs_up(seg.nf)
            )
            matching = 1.0
            for pto in range(1, strong_coupling.order[0]):
                # 0**0=1, from NNLO there is a matching also in this case
                for logpow in range(pto + 1):
                    as_thr = strong_coupling.a(seg.target * xif2, seg.nf - shift + 4)[0]
                    matching += as_thr**pto * L**logpow * m_coeffs[pto, logpow]
            ker_evol *= matching
        ev_mass *= ker_evol
    return m2_ref * ev_mass


def compute(
    masses_ref: HeavyQuarkMasses,
    couplings: CouplingsInfo,
    order: Order,
    evmeth: CouplingEvolutionMethod,
    matching: List[float],
    xif2: float = 1.0,
):
    r"""Compute the |MSbar| masses.

    Computation is performed solving the equation :math:`m_{\bar{MS}}(\mu) =
    \mu` for each heavy quark and consistent boundary contitions.

    Parameters
    ----------
    masses_ref :
        reference scale squared (a.k.a. :math:`Q_{ref}`)
    couplings :
        couplings configuration
    order :
        perturbative order
    evmeth :
        evolution method
    matching :
        threshold matching scale ratios
    xif2 :
        squared ratio of factorization to central scale

    Returns
    -------
    masses : list
        list of |MSbar| masses squared

    """
    # TODO: sketch in the docs how the MSbar computation works with a figure.
    mu2_ref = couplings.scale**2
    nf_ref: FlavorsNumber = couplings.num_flavs_ref
    masses = np.concatenate((np.zeros(nf_ref - 3), np.full(6 - nf_ref, np.inf)))

    def sc(thr_masses):
        return Couplings(
            couplings,
            order=order,
            method=evmeth,
            masses=thr_masses,
            thresholds_ratios=(np.array(matching) * xif2).tolist(),
            hqm_scheme=QuarkMassScheme.MSBAR,
        )

    # First you need to look for the thr around the given as_ref
    heavy_quarks = quark_names[3:]
    hq_idxs = np.arange(0, 3)
    if nf_ref > 4:
        heavy_quarks = reversed(heavy_quarks)
        hq_idxs = reversed(hq_idxs)

    # loop on heavy quarks and compute the msbar masses
    for q_idx, hq in zip(hq_idxs, heavy_quarks):
        mhq: QuarkMassRef = getattr(masses_ref, hq)
        q2m_ref = mhq.scale**2
        m2_ref = mhq.value**2

        # check if mass is already given at the pole -> done
        if q2m_ref == m2_ref:
            masses[q_idx] = m2_ref
            continue

        # update the alphas thr scales
        nf_target = q_idx + 3
        shift = -1

        # check that alphas is given with a consistent number of flavors
        if q_idx + 4 == nf_ref and q2m_ref > mu2_ref:
            raise ValueError(
                f"In MSbar scheme, Qm{hq} should be lower than Qref, "
                f"if alpha_s is given with nfref={nf_ref} at scale Qref={mu2_ref}"
            )
        if q_idx + 4 == nf_ref + 1 and q2m_ref < mu2_ref:
            raise ValueError(
                f"In MSbar scheme, Qm{hq} should be greater than Qref, "
                f"if alpha_s is given with nfref={nf_ref} at scale Qref={mu2_ref}"
            )

        # check that for higher patches you do forward running
        # with consistent conditions
        if q_idx + 3 >= nf_ref and q2m_ref >= m2_ref:
            raise ValueError(
                f"In MSbar scheme, Qm{hq} should be lower than m{hq} "
                f"if alpha_s is given with nfref={nf_ref} at scale Qref={mu2_ref}"
            )

        # check that for lower patches you do backward running
        # with consistent conditions
        if q_idx + 3 < nf_ref:
            if q2m_ref < m2_ref:
                raise ValueError(
                    f"In MSbar scheme, Qm{hq} should be greater than m{hq}"
                    f"if alpha_s is given with nfref={nf_ref} at scale Qref={mu2_ref}"
                )
            nf_target += 1
            shift = 1

        # if the initial condition is not in the target patch,
        # you need to evolve it until nf_target patch wall is reached:
        #   for backward you reach the higher, for forward the lower.
        # len(masses[q2m_ref > masses]) + 3 is the nf at the given reference scale
        nf_ref_cur = len(masses[q2m_ref > masses]) + 3
        if nf_target != nf_ref_cur:
            q2_to = masses[q_idx + shift]
            m2_ref = evolve(
                m2_ref,
                q2m_ref,
                sc(masses),
                matching,
                xif2,
                q2_to,
                nf_ref=nf_ref_cur,
                nf_to=nf_target,
            )
            q2m_ref = q2_to

        # now solve the RGE
        masses[q_idx] = solve(
            m2_ref,
            q2m_ref,
            sc(masses),
            nf_target,
            xif2,
        )

    # Check the msbar ordering
    if not np.allclose(masses, np.sort(masses)):
        raise ValueError("MSbar masses are not to be sorted")
    return np.sort(masses)



"""Caching harmonic sums across :mod:`ekore`."""
from typing import Optional

import numba as nb
import numpy as np
import numpy.typing as npt

from . import w1, w2, w3, w4, w5
from .g_functions import mellin_g3
from .polygamma import recursive_harmonic_sum

# here a register of all possible functions
CACHE_SIZE = 31
_index = iter(range(CACHE_SIZE))
S1 = next(_index)  # = S_1(N)
S2 = next(_index)
S3 = next(_index)
S4 = next(_index)
S5 = next(_index)
Sm1 = next(_index)
Sm2 = next(_index)
Sm3 = next(_index)
Sm4 = next(_index)
Sm5 = next(_index)
S21 = next(_index)
S2m1 = next(_index)
Sm21 = next(_index)
Sm2m1 = next(_index)
S31 = next(_index)
Sm31 = next(_index)
Sm22 = next(_index)
S211 = next(_index)
Sm211 = next(_index)
S1h = next(_index)
S2h = next(_index)
S3h = next(_index)
S1mh = next(_index)
S2mh = next(_index)
S3mh = next(_index)
S1ph = next(_index)
S2ph = next(_index)
S3ph = next(_index)
g3 = next(_index)
S1p2 = next(_index)
g3p2 = next(_index)


@nb.njit(cache=True)
def reset():
    """Return the cache placeholder array."""
    return np.full(CACHE_SIZE, np.nan, np.complex_)


@nb.njit(cache=True)
def update(func, key, cache, n):
    """Compute simple harmonics if not yet in cache."""
    if np.isnan(cache[key]):
        cache[key] = func(n)
    return cache


@nb.njit(cache=True)
def update_Sm1(cache, n, is_singlet):
    """Compute Sm1 if not yet in cache."""
    if np.isnan(cache[Sm1]):
        cache = update(w1.S1, S1, cache, n)
        cache = update(w1.S1, S1mh, cache, (n - 1) / 2)
        cache = update(w1.S1, S1h, cache, n / 2)
        cache[Sm1] = w1.Sm1(n, cache[S1], cache[S1mh], cache[S1h], is_singlet)
    return cache


@nb.njit(cache=True)
def update_Sm2(cache, n, is_singlet):
    """Compute Sm2 if not yet in cache."""
    if np.isnan(cache[Sm2]):
        cache = update(w2.S2, S2, cache, n)
        cache = update(w2.S2, S2mh, cache, (n - 1) / 2)
        cache = update(w2.S2, S2h, cache, n / 2)
        cache[Sm2] = w2.Sm2(n, cache[S2], cache[S2mh], cache[S2h], is_singlet)
    return cache


@nb.njit(cache=True)
def get(
    key: int, cache: npt.ArrayLike, n: complex, is_singlet: Optional[bool] = None
) -> complex:
    r"""Retrieve an element of the cache.

    Parameters
    ----------
    key :
        harmonic sum key
    cache :
        cache list holding all elements
    n :
        complex evaluation point
    is_singlet :
        symmetry factor: True for singlet like quantities
        (:math:`\eta=(-1)^N = 1`),
        False for non-singlet like quantities
        (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    complex :
        requested harmonic sum evaluated at n

    """
    # Maybe improve error
    if key < 0 or key >= len(cache):
        raise RuntimeError
    # load the thing
    s = cache[key]
    # found? i.e. not NaN?
    if not np.isnan(s):
        return s
    # compute it now ...
    # weight 1
    if key == S1:
        s = w1.S1(n)
    elif key == S1h:
        s = w1.S1(n / 2)
    elif key == S1mh:
        s = w1.S1((n - 1) / 2)
    elif key == S1ph:
        cache = update(w1.S1, S1mh, cache, (n - 1) / 2)
        s = recursive_harmonic_sum(cache[S1mh], (n - 1) / 2, 1, 1)
    elif key == Sm1:
        cache = update_Sm1(cache, n, is_singlet)
        s = cache[key]
    elif key == S1p2:
        cache = update(w1.S1, S1, cache, n)
        s = recursive_harmonic_sum(cache[S1], n, 2, 1)
    # weight 2
    elif key == S2:
        s = w2.S2(n)
    elif key == S2h:
        s = w2.S2(n / 2)
    elif key == S2mh:
        s = w2.S2((n - 1) / 2)
    elif key == S2ph:
        cache = update(w2.S2, S2mh, cache, (n - 1) / 2)
        s = recursive_harmonic_sum(cache[S2mh], (n - 1) / 2, 1, 2)
    elif key == Sm2:
        cache = update_Sm2(cache, n, is_singlet)
        s = cache[key]
    # weight 3
    elif key == S3:
        s = w3.S3(n)
    elif key == S3h:
        s = w3.S3(n / 2)
    elif key == S3mh:
        s = w3.S3((n - 1) / 2)
    elif key == S3ph:
        cache = update(w3.S3, S3mh, cache, (n - 1) / 2)
        s = recursive_harmonic_sum(cache[S3mh], (n - 1) / 2, 1, 3)
    elif key == Sm3:
        cache = update(w3.S3, S3, cache, n)
        cache = update(w3.S3, S3mh, cache, (n - 1) / 2)
        cache = update(w3.S3, S3h, cache, n / 2)
        s = w3.Sm3(n, cache[S3], cache[S3mh], cache[S3h], is_singlet)
    # weight 4
    elif key == S4:
        s = w4.S4(n)
    elif key == Sm4:
        cache = update(w4.S4, S4, cache, n)
        S4mh = w4.S4((n - 1) / 2)
        S4h = w4.S4(n / 2)
        s = w4.Sm4(n, cache[S4], S4mh, S4h, is_singlet)
    # weight 5
    elif key == S5:
        s = w5.S5(n)
    elif key == Sm5:
        cache = update(w5.S5, S5, cache, n)
        S5mh = w5.S5((n - 1) / 2)
        S5h = w5.S5(n / 2)
        s = w5.Sm5(n, cache[S5], S5mh, S5h, is_singlet)
    # mellin g3 and related
    elif key == g3:
        cache = update(w1.S1, S1, cache, n)
        s = mellin_g3(n, cache[S1])
    elif key == g3p2:
        cache = update(w1.S1, S1p2, cache, n + 2)
        s = mellin_g3(n + 2, cache[S1p2])
    else:
        # Multi index harmonics which do not require Sxm
        cache = update(w1.S1, S1, cache, n)
        cache = update(w2.S2, S2, cache, n)
        if key == S21:
            s = w3.S21(n, cache[S1], cache[S2])
        elif key == S31:
            cache = update(w3.S3, S3, cache, n)
            cache = update(w4.S4, S4, cache, n)
            s = w4.S31(n, cache[S1], cache[S2], cache[S3], cache[S4])
        elif key == S211:
            cache = update(w3.S3, S3, cache, n)
            s = w4.S211(n, cache[S1], cache[S2], cache[S3])
        else:
            # Multi index harmonics which require Sm1
            cache = update_Sm1(cache, n, is_singlet)
            if key == Sm21:
                s = w3.Sm21(n, cache[S1], cache[Sm1], is_singlet)
            elif key == Sm211:
                s = w4.Sm211(n, cache[S1], cache[S2], cache[Sm1], is_singlet)
            else:
                #  Multi index harmonics which require also Sm2
                cache = update_Sm2(cache, n, is_singlet)
                if key == S2m1:
                    s = w3.S2m1(n, cache[S2], cache[Sm1], cache[Sm2], is_singlet)
                elif key == Sm2m1:
                    s = w3.Sm2m1(n, cache[S1], cache[S2], cache[Sm2])
                elif key == Sm31:
                    s = w4.Sm31(n, cache[S1], cache[Sm1], cache[Sm2], is_singlet)
                elif key == Sm22:
                    if np.isnan(cache[Sm31]):
                        cache[Sm31] = w4.Sm31(
                            n, cache[S1], cache[Sm1], cache[Sm2], is_singlet
                        )
                    s = w4.Sm22(
                        n, cache[S1], cache[S2], cache[Sm2], cache[Sm31], is_singlet
                    )
    # store and return
    cache[key] = s
    return s



"""Types and quantities related to theory couplings."""
import dataclasses
import enum
from typing import Optional

from ..io.dictlike import DictLike
from ..io.types import FlavorsNumber, LinearScale, ReferenceRunning, Scalar

Coupling = Scalar
CouplingRef = ReferenceRunning[Coupling]


@dataclasses.dataclass
class CouplingsInfo(DictLike):
    """Reference values for coupling constants.

    Also includes further information, defining the run of the couplings.

    """

    alphas: Coupling
    alphaem: Coupling
    scale: LinearScale
    max_num_flavs: FlavorsNumber
    num_flavs_ref: Optional[FlavorsNumber]
    r"""Number of active flavors at strong coupling reference scale.

    I.e. :math:`n_{f,\text{ref}}(\mu^2_{\text{ref}})`, formerly called
    ``nfref``.

    """
    em_running: bool = False

    @property
    def values(self):
        """Collect only couplings values."""
        return (self.alphas, self.alphaem)


# TODO: upgrade the following to StrEnum (requires py>=3.11) with that, it is
# possible to replace all non-alias right sides with calls to enum.auto()


class CouplingEvolutionMethod(enum.Enum):
    """Beta functions solution method."""

    EXACT = "exact"
    EXPANDED = "expanded"



"""Weight 5 harmonic sums."""

import numba as nb

from eko.constants import zeta5

from .polygamma import cern_polygamma


@nb.njit(cache=True)
def S5(N):
    r"""Compute the harmonic sum :math:`S_5(N)`.

    .. math::
      S_5(N) = \sum\limits_{j=1}^N \frac 1 {j^5} = \frac 1 24 \psi_4(N+1)+\zeta(5)

    with :math:`\psi_4(N)` the 4th-polygamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    S_5 : complex
        Harmonic sum :math:`S_5(N)`

    See Also
    --------
    ekore.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`

    """
    return zeta5 + 1.0 / 24.0 * cern_polygamma(N + 1.0, 4)


@nb.njit(cache=True)
def Sm5(N, hS5, hS5mh, hS5h, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-5}(N)`.

    .. math::
      S_{-5}(N) = \sum\limits_{j=1}^N \frac {(-1)^j} {j^5}

    Parameters
    ----------
    N : complex
        Mellin moment
    hS5:  complex
        Harmonic sum :math:`S_{5}(N)`
    hS5mh: complex
        Harmonic sum :math:`S_{5}((N-1)/2)`
    hS5h: complex
        Harmonic sum :math:`S_{5}(N/2)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm5 : complex
        Harmonic sum :math:`S_{-5}(N)`

    See Also
    --------
    eko.harmonic.w5.S5 : :math:`S_5(N)`

    """
    if is_singlet is None:
        return (
            1 / 2**4 * ((1 - (-1) ** N) / 2 * hS5mh + ((-1) ** N + 1) / 2 * hS5h)
            - hS5
        )
    if is_singlet:
        return 1 / 2**4 * hS5h - hS5
    return 1 / 2**4 * hS5mh - hS5



"""Collection of QED non-singlet EKOs."""
import numba as nb
import numpy as np

from .. import beta
from . import non_singlet as ns
from . import utils


@nb.njit(cache=True)
def contract_gammas(gamma_ns, aem):
    """Contract anomalous dimension along the QED axis.

    Parameters
    ----------
    gamma_ns : 2D numpy.ndarray
        non-singlet anomalous dimensions
    aem : float
        electromagnetic coupling value

    Returns
    -------
    gamma_ns : 1D numpy.ndarray
        non-singlet anomalous dimensions
    """
    vec_alphaem = np.array(
        [aem**i for i in range(gamma_ns.shape[1])], dtype=np.complex_
    )
    return gamma_ns @ vec_alphaem


@nb.njit(cache=True)
def apply_qed(gamma_pure_qed, mu2_from, mu2_to):
    """Apply pure QED evolution to QCD kernel.

    Parameters
    ----------
    gamma_ns : float
        pure QED part of the AD
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2

    Returns
    -------
    exp : float
        pure QED evolution kernel

    """
    return np.exp(gamma_pure_qed * np.log(mu2_from / mu2_to))


@nb.njit(cache=True)
def as1_exact(gamma_ns, a1, a0, beta):
    """O(as1aem1) non-singlet exact EKO.

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta : list
        list of the values of the beta functions

    Returns
    -------
    e_ns^0 : complex
        O(as1aem1) non-singlet exact EKO
    """
    return ns.lo_exact(gamma_ns, a1, a0, beta)


@nb.njit(cache=True)
def as2_exact(gamma_ns, a1, a0, beta):
    """O(as2aem1) non-singlet exact EKO.

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta : list
        list of the values of the beta functions

    Returns
    -------
    e_ns^1 : complex
        O(as2aem1) non-singlet exact EKO
    """
    return ns.nlo_exact(gamma_ns, a1, a0, beta)


@nb.njit(cache=True)
def as3_exact(gamma_ns, a1, a0, beta):
    """O(as3aem1) non-singlet exact EKO.

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta : list
        list of the values of the beta functions

    Returns
    -------
    e_ns^2 : complex
        O(as3aem1) non-singlet exact EKO
    """
    return ns.nnlo_exact(gamma_ns, a1, a0, beta)


@nb.njit(cache=True)
def dispatcher(
    order,
    method,
    gamma_ns,
    as_list,
    aem_half,
    alphaem_running,
    nf,
    ev_op_iterations,
    mu2_to,
    mu2_from,
):
    r"""Determine used kernel and call it.

    In LO we always use the exact solution.

    Parameters
    ----------
    order : tuple(int,int)
        perturbation order
    method : str
        method
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    aem_list : numpy.ndarray
        electromagnetic coupling values
    alphaem_running : Bool
        running of alphaem
    nf : int
        number of active flavors
    ev_op_iterations : int
        number of evolution steps
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2

    Returns
    -------
    e_ns : complex
        non-singlet EKO
    """
    return exact(
        order,
        gamma_ns,
        as_list,
        aem_half,
        nf,
        ev_op_iterations,
        mu2_to,
        mu2_from,
    )


@nb.njit(cache=True)
def fixed_alphaem_exact(order, gamma_ns, a1, a0, aem, nf, mu2_to, mu2_from):
    """Compute exact solution for fixed alphaem.

    Parameters
    ----------
    order : tuple(int,int)
        perturbation order
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    aem : float
        electromagnetic coupling value
    nf : int
        number of active flavors
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2

    Returns
    -------
    e_ns : complex
        non-singlet EKO
    """
    betalist = [beta.beta_qcd((2 + i, 0), nf) for i in range(order[0])]
    betalist[0] += aem * beta.beta_qcd((2, 1), nf)
    gamma_ns_list = contract_gammas(gamma_ns, aem)
    if order[0] == 1:
        qcd_only = as1_exact(gamma_ns_list[1:], a1, a0, betalist)
    elif order[0] == 2:
        qcd_only = as2_exact(gamma_ns_list[1:], a1, a0, betalist)
    elif order[0] == 3:
        qcd_only = as3_exact(gamma_ns_list[1:], a1, a0, betalist)
    else:
        raise NotImplementedError("Selected order is not implemented")
    return qcd_only * apply_qed(gamma_ns_list[0], mu2_from, mu2_to)


@nb.njit(cache=True)
def exact(order, gamma_ns, as_list, aem_half, nf, ev_op_iterations, mu2_to, mu2_from):
    """Compute exact solution for running alphaem.

    Parameters
    ----------
    order : tuple(int,int)
        perturbation order
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    aem_list : numpy.ndarray
        electromagnetic coupling values
    nf : int
        number of active flavors
    ev_op_iterations : int
        number of evolution steps
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2

    Returns
    -------
    e_ns : complex
        non-singlet EKO
    """
    mu2_steps = utils.geomspace(mu2_from, mu2_to, 1 + ev_op_iterations)
    res = 1.0
    for step in range(1, ev_op_iterations + 1):
        aem = aem_half[step - 1]
        a1 = as_list[step]
        a0 = as_list[step - 1]
        mu2_from = mu2_steps[step - 1]
        mu2_to = mu2_steps[step]
        res *= fixed_alphaem_exact(order, gamma_ns, a1, a0, aem, nf, mu2_to, mu2_from)
    return res



"""Upgrade old files."""
import pathlib
from typing import Optional

import click

from eko.io import legacy

from .base import command


@command.command("convert")
@click.argument("old", type=click.Path(path_type=pathlib.Path, exists=True))
@click.option("-n", "--new", type=click.Path(path_type=pathlib.Path), default=None)
def subcommand(old: pathlib.Path, new: Optional[pathlib.Path]):
    """Convert old EKO files to new format.

    The OLD file path is used also for the new one, appending "-new" to its
    stem, unless it is explicitly specified with the dedicated option.
    """
    if new is None:
        new = old.parent / old.with_stem(old.stem + "-new")

    legacy.load_tar(old, new)



"""Utilities to manipulate unstructured IO.

The content is treated independently on the particular data content, but as
generic uknown data in an abstract file format, e.g. a tar archive or YAML data
file, as opposed to structured YAML representing a specific runcard.

"""
import os
from pathlib import Path
from tarfile import TarFile, TarInfo
from typing import Optional, Sequence


def is_within_directory(directory: os.PathLike, target: os.PathLike) -> bool:
    """Check if target path is contained in directory.

    Thanks to TrellixVulnTeam for the `idea
    <https://github.com/NNPDF/eko/pull/154>`_.

    Parameters
    ----------
    directory:
        the directory where the target is supposed to be contained
    target:
        the target file to check

    """
    abs_dir = Path(directory).absolute()
    abs_target = Path(target).absolute()

    return abs_dir == abs_target or abs_dir in abs_target.parents


def safe_extractall(
    tar: TarFile,
    path: Optional[os.PathLike] = None,
    members: Optional[Sequence[TarInfo]] = None,
    *,
    numeric_owner: bool = False,
):
    """Extract a tar archive avoiding CVE-2007-4559 issue.

    Thanks to TrellixVulnTeam for the `contribution
    <https://github.com/NNPDF/eko/pull/154>`_.

    All undocumented parameters have the same meaning of the analogue ones in
    :meth:`TarFile.extractall`.

    Parameters
    ----------
    tar:
        the tar archive object to be extracted
    path:
        the path to extract to, if not specified the current directory is used

    """
    if path is None:
        path = Path.cwd()
    path = Path(path)

    for member in tar.getmembers():
        member_path = path / member.name
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner=numeric_owner)



"""Logging settings."""
import logging

from rich.console import Console
from rich.logging import RichHandler

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(console=Console(stderr=True))],
)



r"""The polarized, space-like |OME|."""

import numba as nb
import numpy as np

from ....harmonics import cache as c
from . import as1, as2


@nb.njit(cache=True)
def A_singlet(matching_order, n, nf, L):
    r"""Compute the tower of the singlet |OME|.

    Parameters
    ----------
    matching_order : tuple(int,int)
        perturbative matching order
    n : complex
        Mellin variable
    nf: int
        number of active flavor below threshold
    L : float
        :math:``\ln(\mu_F^2 / m_h^2)``

    Returns
    -------
    numpy.ndarray
        singlet |OME|

    """
    A_s = np.zeros((matching_order[0], 3, 3), np.complex_)
    if matching_order[0] >= 1:
        A_s[0] = as1.A_singlet(n, L)
    cache = c.reset()
    if matching_order[0] == 2:
        A_s[1] = as2.A_singlet(n, cache, L, nf)
    if matching_order[0] > 2:
        raise NotImplementedError(
            "Polarized, space-like beyond NNLO is not implemented yet."
        )
    return A_s


@nb.njit(cache=True)
def A_non_singlet(matching_order, n, L):
    r"""Compute the tower of the non-singlet |OME|.

    Parameters
    ----------
    matching_order : tuple(int,int)
        perturbative matching order
    n : complex
        Mellin variable
    L : float
        :math:``\ln(\mu_F^2 / m_h^2)``

    Returns
    -------
    numpy.ndarray
        non-singlet |OME|

    """
    A_ns = np.zeros((matching_order[0], 2, 2), np.complex_)
    cache = c.reset()
    if matching_order[0] == 2:
        A_ns[1] = as2.A_ns(n, cache, L)
    if matching_order[0] > 2:
        raise NotImplementedError(
            "Polarized, space-like beyond NNLO is not implemented yet."
        )
    return A_ns
